#! /usr/bin/env python
# -*- coding: utf-8 -*-
import logging

import numpy as np
import pandas as pd

from saqc.core.reader import readConfig, checkConfig
from saqc.core.config import Fields
from saqc.core.evaluator import evalExpression
from saqc.lib.plotting import plotHook, plotAllHook
from saqc.lib.tools import combineDataFrames
from saqc.flagger import BaseFlagger, CategoricalFlagger, SimpleFlagger, DmpFlagger


def _collectVariables(meta, data):
    """
    find every relevant variable
    """
    # NOTE: get to know every variable from meta
    variables = []
    for idx, configrow in meta.iterrows():
        varname = configrow[Fields.VARNAME]
        # assign = configrow[Fields.ASSIGN]
        if varname in variables:
            continue
        # if (varname in data):  # or (varname not in variables and assign is True):
        variables.append(varname)
    return variables


def _checkInput(data, flags, flagger):
    if not isinstance(data, pd.DataFrame):
        raise TypeError("data must be of type pd.DataFrame")

    if isinstance(data.index, pd.MultiIndex):
        raise TypeError("the index of data is not allowed to be a multiindex")

    if isinstance(data.columns, pd.MultiIndex):
        raise TypeError("the columns of data is not allowed to be a multiindex")

    if not isinstance(flagger, BaseFlagger):
        flaggerlist = [CategoricalFlagger, SimpleFlagger, DmpFlagger]
        raise TypeError(f"flagger must be of type {flaggerlist} or any inherit class from {BaseFlagger}")

    if flags is None:
        return

    if not isinstance(flags, pd.DataFrame):
        raise TypeError("flags must be of type pd.DataFrame")

    if isinstance(data.index, pd.MultiIndex):
        raise TypeError("the index of data is not allowed to be a multiindex")

    if len(data) != len(flags):
        raise ValueError("the index of flags and data has not the same length")

    # NOTE: do not test columns as they not necessarily must be the same


def _handleErrors(exc, configrow, test, policy):
    line = configrow[Fields.LINENUMBER]
    msg = f"config, line {line}, test: '{test}' failed with:\n{type(exc).__name__}: {exc}"
    if policy == "ignore":
        logging.debug(msg)
    elif policy == "warn":
        logging.warning(msg)
    else:
        raise Exception(msg)


def _setup():
    pd.set_option("mode.chained_assignment", "warn")
    np.seterr(invalid="ignore")


def run(
    config_file: str,
    flagger: BaseFlagger,
    data: pd.DataFrame,
    flags: pd.DataFrame = None,
    nodata: float = np.nan,
    error_policy: str = "raise",
) -> (pd.DataFrame, BaseFlagger):

    _setup()
    _checkInput(data, flags, flagger)
    config = readConfig(config_file, data)

    # split config into the test and some 'meta' data
    tests = config.filter(regex=Fields.TESTS)
    meta = config[config.columns.difference(tests.columns)]

    # prepapre the flags
    flag_cols = _collectVariables(meta, data)
    flagger = flagger.initFlags(data=pd.DataFrame(index=data.index, columns=flag_cols))
    if flags is not None:
        flagger = flagger.setFlagger(flagger.initFlags(flags=flags))

    # NOTE:
    # this checks comes late, but the compilation of
    # user-test needs fully prepared flags
    checkConfig(config, data, flagger, nodata)

    # NOTE:
    # the outer loop runs over the flag tests, the inner one over the
    # variables. Switching the loop order would complicate the
    # reference to flags from other variables within the dataset
    for _, testcol in tests.iteritems():

        # NOTE: just an optimization
        if testcol.dropna().empty:
            continue

        for idx, configrow in meta.iterrows():

            # store config params in some handy variables
            varname = configrow[Fields.VARNAME]
            start_date = configrow[Fields.START]
            end_date = configrow[Fields.END]

            func = testcol[idx]
            if pd.isnull(func):
                continue

            if varname not in data and varname not in flagger.getFlags():
                continue

            # NOTE:
            # time slicing support is currently disabled
            # prepare the data for the tests
            # data_chunk = data.loc[start_date:end_date]
            data_chunk = data
            if data_chunk.empty:
                continue
            flagger_chunk = flagger.getFlagger(loc=data_chunk.index)

            try:
                # actually run the tests
                data_chunk_result, flagger_chunk_result = evalExpression(
                    func, data=data_chunk, field=varname, flagger=flagger_chunk, nodata=nodata,
                )
            except Exception as e:
                _handleErrors(e, configrow, func, error_policy)
                continue

            if configrow[Fields.PLOT]:
                plotHook(
                    data_chunk_result, flagger_chunk, flagger_chunk_result, varname, func,
                )

            # NOTE:
            # time slicing support is currently disabled
            # flagger = flagger.setFlagger(flagger_chunk_result)
            # data = combineDataFrames(data, data_chunk_result)
            flagger = flagger_chunk_result
            data = data_chunk_result

    plotAllHook(data, flagger)

    return data, flagger



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import click

import numpy as np
import pandas as pd

from saqc.core import run
from saqc.flagger import CategoricalFlagger


FLAGGERS = {
    "numeric": CategoricalFlagger([-1, 0, 1]),
    "category": CategoricalFlagger(["NIL", "OK", "BAD"]),
}


@click.command()
@click.option(
    "-c", "--config", type=click.Path(exists=True), required=True, help="path to the configuration file",
)
@click.option(
    "-d", "--data", type=click.Path(exists=True), required=True, help="path to the data file",
)
@click.option("-o", "--outfile", type=click.Path(exists=False), help="path to the output file")
@click.option(
    "--flagger", default="category", type=click.Choice(FLAGGERS.keys()), help="the flagging scheme to use",
)
@click.option("--nodata", default=np.nan, help="nodata value")
@click.option("--fail/--no-fail", default=True, help="whether to stop the program run on errors")
def main(config, data, flagger, outfile, nodata, fail):

    data = pd.read_csv(data, index_col=0, parse_dates=True,)

    data_result, flagger_result = run(
        config_file=config,
        flagger=FLAGGERS[flagger],
        data=data,
        nodata=nodata,
        error_policy="raise" if fail else "warn",
    )

    if outfile:
        flags = flagger_result.getFlags()
        flags_out = flags.where((flags.isnull() | flagger_result.isFlagged()), flagger_result.GOOD)
        cols_out = sum([[c, c + "_flags"] for c in flags_out], [])
        data_out = data_result.join(flags_out, rsuffix="_flags")
        data_out[cols_out].to_csv(outfile, header=True, index=True)


if __name__ == "__main__":
    main()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import pytest

from saqc.funcs.break_detection import flagBreaksSpektrumBased
from test.common import TESTFLAGGER, initData


@pytest.fixture
def data():
    return initData(
        1, start_date="2011-01-01 00:00:00", end_date="2011-01-02 03:00:00", freq="5min"
    )


@pytest.mark.parametrize("flagger", TESTFLAGGER)
def test_flagBreaks_SpektrumBased(data, flagger):
    field, *_ = data.columns
    data.iloc[5:15] += 100
    break_positions = [5, 15]
    flagger = flagger.initFlags(data)
    data, flagger_result = flagBreaksSpektrumBased(data, field, flagger)
    flag_result = flagger_result.getFlags(field)
    test_sum = (flag_result[break_positions] == flagger.BAD).sum()
    assert test_sum == len(break_positions)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import pandas as pd

from saqc.lib.tools import sesonalMask, flagWindow, groupConsecutives

from saqc.funcs.register import register


@register("flagGeneric")
def flagGeneric(data, field, flagger, func, func_arguments, **kwargs):
    # NOTE:
    # The naming of the func parameter is pretty confusing
    # as it actually holds the result of a generic expression
    mask = func.squeeze()
    if np.isscalar(mask):
        raise TypeError(f"generic expression does not return an array")
    if not np.issubdtype(mask.dtype, np.bool_):
        raise TypeError(f"generic expression does not return a boolean array")
    flagger = flagger.setFlags(field, mask, **kwargs)
    return data, flagger


@register("flagWindowAfterFlag")
def flagWindowAfterFlag(data, field, flagger, window, func, **kwargs):
    data, flagger_new = func
    flagger_repeated = flagWindow(flagger, flagger_new, field, direction="fw", window=window, **kwargs)
    return data, flagger_repeated


@register("flagNextAfterFlag")
def flagNextAfterFlag(data, field, flagger, n, func, **kwargs):
    data, flagger_new = func
    flagger_repeated = flagWindow(flagger, flagger_new, field, direction="fw", window=n, **kwargs)
    return data, flagger_repeated


@register("range")
def flagRange(data, field, flagger, min, max, **kwargs):
    datacol = data[field].values
    mask = (datacol < min) | (datacol > max)
    flagger = flagger.setFlags(field, mask, **kwargs)
    return data, flagger


@register("missing")
def flagMissing(data, field, flagger, nodata=np.nan, **kwargs):
    datacol = data[field]
    if np.isnan(nodata):
        mask = datacol.isna()
    else:
        mask = datacol[datacol == nodata]

    flagger = flagger.setFlags(field, loc=mask, **kwargs)
    return data, flagger


@register("sesonalRange")
def flagSesonalRange(
    data, field, flagger, min, max, startmonth=1, endmonth=12, startday=1, endday=31, **kwargs,
):
    smask = sesonalMask(data.index, startmonth, startday, endmonth, endday)

    d = data.loc[smask, [field]]
    if d.empty:
        return data, flagger

    _, flagger_range = flagRange(d, field, flagger.getFlagger(loc=d.index), min=min, max=max, **kwargs)

    if not flagger_range.isFlagged(field).any():
        return data, flagger

    flagger = flagger.setFlagger(flagger_range)
    return data, flagger


@register("clear")
def clearFlags(data, field, flagger, **kwargs):
    flagger = flagger.clearFlags(field, **kwargs)
    return data, flagger


@register("force")
def forceFlags(data, field, flagger, flag, **kwargs):
    flagger = flagger.clearFlags(field).setFlags(field, flag=flag, **kwargs)
    return data, flagger


@register("isolated")
def flagIsolated(
    data, field, flagger, gap_window, group_window, **kwargs,
):

    gap_window = pd.tseries.frequencies.to_offset(gap_window)
    group_window = pd.tseries.frequencies.to_offset(group_window)

    col = data[field].mask(flagger.isFlagged(field))
    mask = col.isnull()

    flags = pd.Series(data=0, index=col.index, dtype=bool)
    for srs in groupConsecutives(mask):
        if np.all(~srs):
            start = srs.index[0]
            stop = srs.index[-1]
            if stop - start <= group_window:
                left = mask[start - gap_window : start].iloc[:-1]
                if left.all():
                    right = mask[stop : stop + gap_window].iloc[1:]
                    if right.all():
                        flags[start:stop] = True

    flagger = flagger.setFlags(field, flags, **kwargs)

    return data, flagger


@register("dummy")
def flagDummyFunction(data, field, flagger, **kwargs):
    return data, flagger



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from saqc.flagger.categoricalflagger import CategoricalFlagger


FLAGS = [-1, 0, 1]


class SimpleFlagger(CategoricalFlagger):
    def __init__(self):
        super().__init__(FLAGS)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import operator as op
from copy import deepcopy
from collections import OrderedDict
from abc import ABC, abstractmethod
from typing import TypeVar, Union, Any

import numpy as np
import pandas as pd

from saqc.lib.tools import toSequence, assertScalar, assertDataFrame


COMPARATOR_MAP = {
    "!=": op.ne,
    "==": op.eq,
    ">=": op.ge,
    ">": op.gt,
    "<=": op.le,
    "<": op.lt,
}


BaseFlaggerT = TypeVar("BaseFlaggerT")
PandasT = Union[pd.Series, pd.DataFrame]
# TODO: get some real types here (could be tricky...)
LocT = Any
IlocT = Any
FlagT = Any


class BaseFlagger(ABC):
    @abstractmethod
    def __init__(self, dtype):
        # NOTE: the type of the _flags DataFrame
        self.dtype = dtype
        # NOTE: the arggumens of setFlags supported from
        #       the configuration functions
        self.signature = ("flag",)
        self._flags: pd.DataFrame

    def initFlags(self, data: pd.DataFrame = None, flags: pd.DataFrame = None) -> BaseFlaggerT:
        """
        initialize a flagger based on the given 'data' or 'flags'
        if 'data' is not None: return a flagger with flagger.UNFALGGED values
        if 'flags' is not None: return a flagger with the given flags
        """

        if data is None and flags is None:
            raise TypeError("either 'data' or 'flags' are required")
        if data is not None:
            flags = pd.DataFrame(data=self.UNFLAGGED, index=data.index, columns=data.columns)
        return self._copy(self._assureDtype(flags))

    def setFlagger(self, other: BaseFlaggerT):
        """
        Merge the given flagger 'other' into self
        """
        # NOTE: add more checks !?
        if not isinstance(other, self.__class__):
            raise TypeError(f"flagger of type '{self.__class__}' needed")

        this = self._flags
        other = other._flags

        flags = this.reindex(
            index=this.index.union(other.index),
            columns=this.columns.union(other.columns, sort=False),
            fill_value=self.UNFLAGGED,
        )

        for key, values in other.iteritems():
            flags.loc[other.index, key] = values

        return self._copy(self._assureDtype(flags))

    def getFlagger(self, field: str = None, loc: LocT = None, iloc: IlocT = None) -> BaseFlaggerT:
        """
        return a potentially trimmed down copy of self
        """
        assertScalar("field", field, optional=True)
        mask = self._locatorMask(field=slice(None), loc=loc, iloc=iloc)
        flags = self._flags.loc[mask, field or slice(None)]
        if isinstance(flags, pd.Series):
            flags = flags.to_frame()
        return self._copy(flags)

    def getFlags(self, field: str = None, loc: LocT = None, iloc: IlocT = None) -> PandasT:
        """
        return a copy of a potentially trimmed down 'self._flags' DataFrame
        """
        assertScalar("field", field, optional=True)
        field = field or slice(None)
        flags = self._flags.copy()
        mask = self._locatorMask(field, loc, iloc)
        return flags.loc[mask, field]

    def setFlags(
        self, field: str, loc: LocT = None, iloc: IlocT = None, flag: FlagT = None, force: bool = False, **kwargs,
    ) -> BaseFlaggerT:
        assertScalar("field", field, optional=False)

        flag = self.BAD if flag is None else self._checkFlag(flag)

        this = self.getFlags(field=field)
        other = self._broadcastFlags(field=field, flag=flag)

        mask = self._locatorMask(field, loc, iloc)
        if not force:
            mask &= (this < other).values

        out = deepcopy(self)
        out._flags.loc[mask, field] = other[mask]
        return out

    def clearFlags(self, field: str, loc: LocT = None, iloc: IlocT = None, **kwargs) -> BaseFlaggerT:
        assertScalar("field", field, optional=False)
        return self.setFlags(field=field, loc=loc, iloc=iloc, flag=self.UNFLAGGED, force=True, **kwargs)

    def isFlagged(
        self, field=None, loc: LocT = None, iloc: IlocT = None, flag: FlagT = None, comparator: str = ">", **kwargs,
    ) -> PandasT:
        assertScalar("field", field, optional=True)
        assertScalar("flag", flag, optional=True)
        self._checkFlag(flag)
        flag = self.GOOD if flag is None else flag
        flags = self.getFlags(field, loc, iloc, **kwargs)
        cp = COMPARATOR_MAP[comparator]
        flagged = pd.notna(flags) & cp(flags, flag)
        return flagged

    def _copy(self, flags: pd.DataFrame = None) -> BaseFlaggerT:
        out = deepcopy(self)
        if flags is not None:
            out._flags = flags
        return out

    def _locatorMask(self, field: str = None, loc: LocT = None, iloc: IlocT = None) -> PandasT:
        field = field or slice(None)
        locator = [l for l in (loc, iloc, slice(None)) if l is not None][0]
        index = self._flags.index
        mask = pd.Series(data=np.zeros(len(index), dtype=bool), index=index)
        mask[locator] = True
        return mask

    def _broadcastFlags(self, field: str, flag: FlagT) -> pd.Series:

        this = self.getFlags(field)

        if np.isscalar(flag):
            flag = np.full_like(this, flag)

        return pd.Series(data=flag, index=this.index, name=field, dtype=self.dtype)

    def _checkFlag(self, flag):
        if flag is not None and not self._isDtype(flag):
            raise TypeError(f"invalid flag value '{flag}' for flagger 'self.__class__'")
        return flag

    def _assureDtype(self, flags):
        # NOTE: building up new DataFrames is significantly
        #       faster than assigning into existing ones
        if isinstance(flags, pd.Series):
            return flags.astype(self.dtype)
        tmp = OrderedDict()
        for c in flags.columns:
            tmp[c] = flags[c].astype(self.dtype)
        return pd.DataFrame(tmp)

    @abstractmethod
    def _isDtype(self, flag) -> bool:
        pass

    @property
    @abstractmethod
    def UNFLAGGED(self) -> FlagT:
        """ Return the flag that indicates unflagged data """
        pass

    @property
    @abstractmethod
    def GOOD(self) -> FlagT:
        """ Return the flag that indicates the very best data """
        pass

    @property
    @abstractmethod
    def BAD(self) -> FlagT:
        """ Return the flag that indicates the worst data """
        pass

    @abstractmethod
    def isSUSPICIOUS(self, flag: FlagT) -> bool:
        """ Return bool that indicates if the given flag is valid, but neither
        UNFLAGGED, BAD, nor GOOD."""
        pass



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import pytest
import numpy as np
import pandas as pd

from saqc.funcs import register, flagRange
from saqc.core.core import run
from saqc.core.config import Fields as F
from saqc.lib.plotting import _plot
from test.common import initData, initMetaDict, TESTFLAGGER


OPTIONAL = [False, True]


@register("flagAll")
def flagAll(data, field, flagger, **kwargs):
    # NOTE: remember to rename flag -> flag_values
    return data, flagger.setFlags(field=field, flag=flagger.BAD)


@pytest.fixture
def data():
    return initData(3)


def _initFlags(flagger, data, optional):
    return None
    if optional:
        return flagger.initFlags(data[data.columns[::2]])._flags


@pytest.fixture
def flags(flagger, data, optional):
    if not optional:
        return flagger.initFlags(data[data.columns[::2]])._flags


# NOTE: there is a lot of pytest magic involved:
#       the parametrize parameters are implicitly available
#       within the used fixtures, that is why we need the optional
#       parametrization without actually using it in the
#       function
@pytest.mark.skip(reason="test slicing support is currently disabled")
@pytest.mark.parametrize("flagger", TESTFLAGGER)
@pytest.mark.parametrize("optional", OPTIONAL)
def test_temporalPartitioning(data, flagger, flags):
    """
    Check if the time span in meta is respected
    """
    var1, var2, var3, *_ = data.columns
    split_date = data.index[len(data.index) // 2]

    metadict = [
        {F.VARNAME: var1, F.TESTS: "flagAll()"},
        {F.VARNAME: var2, F.TESTS: "flagAll()", F.END: split_date},
        {F.VARNAME: var3, F.TESTS: "flagAll()", F.START: split_date},
    ]
    meta_file, meta_frame = initMetaDict(metadict, data)
    pdata, pflagger = run(meta_file, flagger, data, flags=flags)

    fields = [F.VARNAME, F.START, F.END]
    for _, row in meta_frame.iterrows():
        vname, start_date, end_date = row[fields]
        fchunk = pflagger.getFlags(field=vname, loc=pflagger.isFlagged(vname))
        assert fchunk.index.min() == start_date, "different start dates"
        assert fchunk.index.max() == end_date, "different end dates"


@pytest.mark.skip(reason="test slicing support is currently disabled")
@pytest.mark.parametrize("flagger", TESTFLAGGER)
@pytest.mark.parametrize("optional", OPTIONAL)
def test_positionalPartitioning(data, flagger, flags):
    data = data.reset_index(drop=True)
    if flags is not None:
        flags = flags.reset_index(drop=True)
    var1, var2, var3, *_ = data.columns
    split_index = int(len(data.index) // 2)

    metadict = [
        {F.VARNAME: var1, F.TESTS: "flagAll()"},
        {F.VARNAME: var2, F.TESTS: "flagAll()", F.END: split_index},
        {F.VARNAME: var3, F.TESTS: "flagAll()", F.START: split_index},
    ]
    meta_file, meta_frame = initMetaDict(metadict, data)

    pdata, pflagger = run(meta_file, flagger, data, flags=flags)

    fields = [F.VARNAME, F.START, F.END]
    for _, row in meta_frame.iterrows():
        vname, start_index, end_index = row[fields]
        fchunk = pflagger.getFlags(field=vname, loc=pflagger.isFlagged(vname))
        assert fchunk.index.min() == start_index, "different start indices"
        assert (
            fchunk.index.max() == end_index
        ), f"different end indices: {fchunk.index.max()} vs. {end_index}"


@pytest.mark.parametrize("flagger", TESTFLAGGER)
@pytest.mark.parametrize("optional", OPTIONAL)
def test_missingConfig(data, flagger, flags):
    """
    Test if variables available in the dataset but not the config
    are handled correctly, i.e. are ignored
    """
    var1, var2, *_ = data.columns

    metadict = [{F.VARNAME: var1, F.TESTS: "flagAll()"}]
    metafobj, meta = initMetaDict(metadict, data)

    pdata, pflagger = run(metafobj, flagger, data, flags=flags)

    assert var1 in pdata and var2 not in pflagger.getFlags()


@pytest.mark.parametrize("flagger", TESTFLAGGER)
def test_errorHandling(data, flagger):

    @register("raisingFunc")
    def _raisingFunc(data, fielf, flagger, **kwargs):
        raise TypeError

    var1, *_ = data.columns

    metadict = [
        {F.VARNAME: var1, F.TESTS: "raisingFunc()"},
    ]

    tests = [
        "ignore",
        "warn"
    ]

    for policy in tests:
        # NOTE: should not fail, that's all we are testing here
        metafobj, _ = initMetaDict(metadict, data)
        run(metafobj, flagger, data, error_policy=policy)


@pytest.mark.parametrize("flagger", TESTFLAGGER)
def test_duplicatedVariable(flagger):
    data = initData(1)
    var1, *_ = data.columns

    metadict = [
        {F.VARNAME: var1, F.TESTS: "flagAll()"},
        {F.VARNAME: var1, F.TESTS: "flagAll()"},
    ]
    metafobj, meta = initMetaDict(metadict, data)

    pdata, pflagger = run(metafobj, flagger, data)
    pflags = pflagger.getFlags()

    if isinstance(pflags.columns, pd.MultiIndex):
        cols = pflags.columns.get_level_values(0).drop_duplicates()
        assert np.all(cols == [var1])
    else:
        assert (pflags.columns == [var1]).all()


@pytest.mark.parametrize("flagger", TESTFLAGGER)
def test_assignVariable(flagger):
    """
    test implicit assignments
    """
    data = initData(1)
    var1, *_ = data.columns
    var2 = "empty"

    metadict = [
        {F.VARNAME: var1, F.TESTS: "flagAll()"},
        {F.VARNAME: var2, F.TESTS: "flagAll()"},
    ]
    metafobj, meta = initMetaDict(metadict, data)

    pdata, pflagger = run(metafobj, flagger, data)
    pflags = pflagger.getFlags()

    if isinstance(pflags.columns, pd.MultiIndex):
        cols = pflags.columns.get_level_values(0).drop_duplicates()
        assert (cols == [var1, var2]).all()
        assert pflagger.isFlagged(var2).any()
    else:
        assert (pflags.columns == [var1, var2]).all()
        assert pflagger.isFlagged(var2).any()


@pytest.mark.parametrize("flagger", TESTFLAGGER)
@pytest.mark.parametrize("optional", OPTIONAL)
def test_dtypes(data, flagger, flags):
    """
    Test if the categorical dtype is preserved through the core functionality
    """
    flagger = flagger.initFlags(data)
    flags = flagger.getFlags()
    var1, var2, *_ = data.columns

    metadict = [
        {F.VARNAME: var1, F.TESTS: "flagAll()"},
        {F.VARNAME: var2, F.TESTS: "flagAll()"},
    ]
    metafobj, meta = initMetaDict(metadict, data)
    pdata, pflagger = run(metafobj, flagger, data, flags=flags)
    pflags = pflagger.getFlags()
    assert dict(flags.dtypes) == dict(pflags.dtypes)


@pytest.mark.parametrize("flagger", TESTFLAGGER)
def test_plotting(data, flagger):
    """
    Test if the plotting code runs, does not show any plot.

    NOTE:
    This test is ignored if matplotlib is not available on the test-system
    """
    pytest.importorskip("matplotlib", reason="requires matplotlib")
    field, *_ = data.columns
    flagger = flagger.initFlags(data)
    _, flagger_range = flagRange(data, field, flagger, min=10, max=90, flag=flagger.BAD)
    _, flagger_range = flagRange(
        data, field, flagger_range, min=40, max=60, flag=flagger.GOOD
    )
    mask = flagger.getFlags(field) != flagger_range.getFlags(field)
    _plot(data, mask, field, flagger, interactive_backend=False)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import pytest

import numpy as np
import pandas as pd

from test.common import TESTFLAGGER

from saqc.funcs.harm_functions import (
    harmonize,
    deharmonize,
    _interpolate,
    _interpolateGrid,
    _insertGrid,
    _outsortCrap,
    linear2Grid,
    interpolate2Grid,
    shift2Grid,
    aggregate2Grid,
    downsample
)


RESHAPERS = ["nshift", "fshift", "bshift"]

COFLAGGING = [False, True]

SETSHIFTCOMMENT = [False, True]

INTERPOLATIONS = ["fshift", "bshift", "nshift", "nagg", "bagg"]

INTERPOLATIONS2 = ["fagg", "time", "polynomial"]

FREQS = ["15min", "30min"]


@pytest.fixture
def data():
    index = pd.date_range(
        start="1.1.2011 00:00:00", end="1.1.2011 01:00:00", freq="15min"
    )
    index = index.insert(2, pd.Timestamp(2011, 1, 1, 0, 29, 0))
    index = index.insert(2, pd.Timestamp(2011, 1, 1, 0, 28, 0))
    index = index.insert(5, pd.Timestamp(2011, 1, 1, 0, 32, 0))
    index = index.insert(5, pd.Timestamp(2011, 1, 1, 0, 31, 0))
    index = index.insert(0, pd.Timestamp(2010, 12, 31, 23, 57, 0))
    index = index.drop(pd.Timestamp("2011-01-01 00:30:00"))
    dat = pd.Series(np.linspace(-50, 50, index.size), index=index, name="data")
    # good to have some nan
    dat[-3] = np.nan
    data = dat.to_frame()
    return data


@pytest.fixture
def multi_data():
    index = pd.date_range(
        start="1.1.2011 00:00:00", end="1.1.2011 01:00:00", freq="15min"
    )
    index = index.insert(2, pd.Timestamp(2011, 1, 1, 0, 29, 0))
    index = index.insert(2, pd.Timestamp(2011, 1, 1, 0, 28, 0))
    index = index.insert(5, pd.Timestamp(2011, 1, 1, 0, 32, 0))
    index = index.insert(5, pd.Timestamp(2011, 1, 1, 0, 31, 0))
    index = index.insert(0, pd.Timestamp(2010, 12, 31, 23, 57, 0))
    index = index.drop(pd.Timestamp("2011-01-01 00:30:00"))
    dat = pd.Series(np.linspace(-50, 50, index.size), index=index, name="data")
    # good to have some nan
    dat[-3] = np.nan
    data = dat.to_frame()
    data.index = data.index.shift(1, "2min")
    dat2 = data.copy()
    dat2.index = dat2.index.shift(1, "17min")
    dat2.rename(columns={"data": "data2"}, inplace=True)
    dat3 = data.copy()
    dat3.index = dat3.index.shift(1, "1h")
    dat3.rename(columns={"data": "data3"}, inplace=True)
    dat3.drop(dat3.index[2:-2], inplace=True)
    # merge
    data = pd.merge(data, dat2, how="outer", left_index=True, right_index=True)
    data = pd.merge(data, dat3, how="outer", left_index=True, right_index=True)
    return data


@pytest.mark.parametrize("flagger", TESTFLAGGER)
@pytest.mark.parametrize("reshaper", RESHAPERS)
@pytest.mark.parametrize("co_flagging", COFLAGGING)
def test_harmSingleVarIntermediateFlagging(data, flagger, reshaper, co_flagging):

    flagger = flagger.initFlags(data)
    # flags = flagger.initFlags(data)
    # make pre harm copies:
    pre_data = data.copy()
    pre_flags = flagger.getFlags()
    freq = "15min"

    # harmonize data:
    data, flagger = harmonize(data, "data", flagger, freq, "time", reshaper)

    # flag something bad
    flagger = flagger.setFlags("data", loc=data.index[3:4])
    data, flagger = deharmonize(data, "data", flagger, co_flagging=co_flagging)

    if reshaper is "nshift":
        if co_flagging is True:
            assert flagger.isFlagged(loc=data.index[3:7]).squeeze().all()
            assert (~flagger.isFlagged(loc=data.index[0:3]).squeeze()).all()
            assert (~flagger.isFlagged(loc=data.index[7:]).squeeze()).all()
        if co_flagging is False:
            assert (
                flagger.isFlagged().squeeze()
                == [False, False, False, False, True, False, True, False, False]
            ).all()
    if reshaper is "bshift":
        if co_flagging is True:
            assert flagger.isFlagged(loc=data.index[5:7]).squeeze().all()
            assert (~flagger.isFlagged(loc=data.index[0:5]).squeeze()).all()
            assert (~flagger.isFlagged(loc=data.index[7:]).squeeze()).all()
        if co_flagging is False:
            assert (
                flagger.isFlagged().squeeze()
                == [False, False, False, False, False, True, True, False, False]
            ).all()
    if reshaper is "fshift":
        if co_flagging is True:
            assert flagger.isFlagged(loc=data.index[3:5]).squeeze().all()
            assert flagger.isFlagged(loc=data.index[6:7]).squeeze().all()
            assert (~flagger.isFlagged(loc=data.index[0:3]).squeeze()).all()
            assert (~flagger.isFlagged(loc=data.index[7:]).squeeze()).all()
        if co_flagging is False:
            assert (
                flagger.isFlagged().squeeze()
                == [False, False, False, False, True, False, True, False, False]
            ).all()

    flags = flagger.getFlags()
    assert pre_data.equals(data)
    assert len(data) == len(flags)
    assert (pre_flags.index == flags.index).all()


@pytest.mark.parametrize("flagger", TESTFLAGGER)
@pytest.mark.parametrize("interpolation", INTERPOLATIONS)
@pytest.mark.parametrize("freq", FREQS)
def test_harmSingleVarInterpolations(data, flagger, interpolation, freq):
    flagger = flagger.initFlags(data)
    flags = flagger.getFlags()
    # make pre harm copies:
    pre_data = data.copy()
    pre_flags = flags.copy()

    harm_start = data.index[0].floor(freq=freq)
    harm_end = data.index[-1].ceil(freq=freq)
    test_index = pd.date_range(start=harm_start, end=harm_end, freq=freq)
    data, flagger = harmonize(
        data,
        "data",
        flagger,
        freq,
        interpolation,
        "fshift",
        reshape_shift_comment=False,
        inter_agg="sum",
    )

    if interpolation is "fshift":
        if freq == "15min":
            assert data.equals(
                pd.DataFrame(
                    {"data": [np.nan, -37.5, -25.0, 0.0, 37.5, 50.0]}, index=test_index
                )
            )
        if freq == "30min":
            assert data.equals(
                pd.DataFrame({"data": [np.nan, -37.5, 0.0, 50.0]}, index=test_index)
            )
    if interpolation is "bshift":
        if freq == "15min":
            assert data.equals(
                pd.DataFrame(
                    {"data": [-50.0, -37.5, -25.0, 12.5, 37.5, 50.0]}, index=test_index
                )
            )
        if freq == "30min":
            assert data.equals(
                pd.DataFrame({"data": [-50.0, -37.5, 12.5, 50.0]}, index=test_index)
            )
    if interpolation is "nshift":
        if freq == "15min":
            assert data.equals(
                pd.DataFrame(
                    {"data": [np.nan, -37.5, -25.0, 12.5, 37.5, 50.0]}, index=test_index
                )
            )
        if freq == "30min":
            assert data.equals(
                pd.DataFrame({"data": [np.nan, -37.5, 12.5, 50.0]}, index=test_index)
            )
    if interpolation is "nagg":
        if freq == "15min":
            assert data.equals(
                pd.DataFrame(
                    {"data": [np.nan, -87.5, -25.0, 0.0, 37.5, 50.0]}, index=test_index
                )
            )
        if freq == "30min":
            assert data.equals(
                pd.DataFrame({"data": [np.nan, -87.5, -25.0, 87.5]}, index=test_index)
            )
    if interpolation is "bagg":
        if freq == "15min":
            assert data.equals(
                pd.DataFrame(
                    {"data": [-50.0, -37.5, -37.5, 12.5, 37.5, 50.0]}, index=test_index
                )
            )
        if freq == "30min":
            assert data.equals(
                pd.DataFrame({"data": [-50.0, -75.0, 50.0, 50.0]}, index=test_index)
            )

    data, flagger = deharmonize(data, "data", flagger, co_flagging=True)

    #data, flagger = deharmonize(data, "data", flagger, co_flagging=True)
    flags = flagger.getFlags()

    assert pre_data.equals(data)
    assert len(data) == len(flags)
    assert (pre_flags.index == flags.index).all()


@pytest.mark.parametrize("flagger", TESTFLAGGER)
@pytest.mark.parametrize("shift_comment", SETSHIFTCOMMENT)
def test_multivariatHarmonization(multi_data, flagger, shift_comment):
    flagger = flagger.initFlags(multi_data)
    flags = flagger.getFlags()
    # for comparison
    pre_data = multi_data.copy()
    pre_flags = flags.copy()
    freq = "15min"

    harm_start = multi_data.index[0].floor(freq=freq)
    harm_end = multi_data.index[-1].ceil(freq=freq)
    test_index = pd.date_range(start=harm_start, end=harm_end, freq=freq)
    # harm:
    multi_data, flagger = harmonize(
        multi_data,
        "data",
        flagger,
        freq,
        "time",
        "nshift",
        reshape_shift_comment=shift_comment,
    )

    multi_data, flagger = harmonize(
        multi_data,
        "data2",
        flagger,
        freq,
        "bagg",
        "bshift",
        inter_agg="sum",
        reshape_agg="max",
        reshape_shift_comment=shift_comment,
    )

    multi_data, flagger = harmonize(
        multi_data,
        "data3",
        flagger,
        freq,
        "fshift",
        "fshift",
        reshape_shift_comment=shift_comment,
    )
    assert multi_data.index.equals(test_index)
    assert pd.Timedelta(pd.infer_freq(multi_data.index)) == pd.Timedelta(freq)

    multi_data, flagger = deharmonize(multi_data, "data3", flagger, co_flagging=False)
    multi_data, flagger = deharmonize(multi_data, "data2", flagger, co_flagging=True)
    multi_data, flagger = deharmonize(multi_data, "data", flagger, co_flagging=True)

    flags = flagger.getFlags()
    assert pre_data.equals(multi_data[pre_data.columns.to_list()])
    assert len(multi_data) == len(flags)
    assert (pre_flags.index == flags.index).all()


@pytest.mark.parametrize("method", INTERPOLATIONS2)
def test_gridInterpolation(data, method):
    freq = "15min"
    data = (data * np.sin(data)).append(data.shift(1, "2h")).shift(1, "3s")
    data = data.squeeze()
    # we are just testing if the interpolation gets passed to the series without causing an error:
    _interpolateGrid(
        data, freq, method, order=1, agg_method="sum", downcast_interpolation=True
    )
    if method == "polynomial":
        _interpolateGrid(
            data, freq, method, order=2, agg_method="sum", downcast_interpolation=True
        )
        _interpolateGrid(
            data, freq, method, order=10, agg_method="sum", downcast_interpolation=True
        )
        data = _insertGrid(data, freq)
        _interpolate(data, method, inter_limit=3)


@pytest.mark.parametrize("flagger", TESTFLAGGER)
def test_outsortCrap(data, flagger):

    field = data.columns[0]
    flagger = flagger.initFlags(data)
    flagger = flagger.setFlags(field, iloc=slice(5, 7))

    drop_index = data.index[5:7]
    d, _ = _outsortCrap(data, field, flagger, drop_flags=flagger.BAD)
    assert drop_index.difference(d.index).equals(drop_index)

    flagger = flagger.setFlags(field, iloc=slice(0, 1), flag=flagger.GOOD)
    drop_index = drop_index.insert(-1, data.index[0])
    d, _ = _outsortCrap(data, field, flagger, drop_flags=[flagger.BAD, flagger.GOOD],)
    assert drop_index.sort_values().difference(d.index).equals(drop_index.sort_values())

    f_drop, _ = _outsortCrap(
        data, field, flagger, drop_flags=[flagger.BAD, flagger.GOOD], return_drops=True,
    )
    assert f_drop.index.sort_values().equals(drop_index.sort_values())

@pytest.mark.parametrize("flagger", TESTFLAGGER)
def test_wrapper(data, flagger):
    # we are only testing, whether the wrappers do pass processing:
    field = data.columns[0]
    freq = '15min'
    flagger = flagger.initFlags(data)
    downsample(data, field, flagger, '15min', '30min', agg_func="sum", sample_func="mean")
    linear2Grid(data, field, flagger, freq, method='nagg', func="max", drop_flags=None)
    aggregate2Grid(data, field, flagger, freq, value_func="sum",
                   flag_func="max", method='nagg', drop_flags=None)
    shift2Grid(data, field, flagger, freq, method='nshift', drop_flags=None)
    interpolate2Grid(data, field, flagger, freq, method="spline")




#! /usr/bin/env python
# -*- coding: utf-8 -*-

import pytest
import numpy as np

from saqc.core.reader import checkConfig
from saqc.core.config import Fields as F
from test.common import initData, initMetaDict, initMetaString, TESTFLAGGER, TESTNODATA


@pytest.fixture
def data():
    return initData(3)


def test_configPreparation(data):
    var1, var2, var3, *_ = data.columns
    date = data.index[len(data.index) // 2]

    # NOTE:
    # time slicing support is currently disabled
    tests = [
        # {F.VARNAME: var1, F.START: date, F.TESTS: "flagAll()", F.PLOT: True},
        {F.VARNAME: var2, F.TESTS: "flagAll()", F.PLOT: False},
        # {F.VARNAME: var3, F.END: date, F.TESTS: "flagAll()"},
        {F.VARNAME: var3, F.TESTS: "flagAll()",},
    ]

    defaults = {
        F.START: data.index.min(),
        F.END: data.index.max(),
        F.PLOT: False,
        F.LINENUMBER: 2,
    }

    for i, test in enumerate(tests):
        _, meta_frame = initMetaDict([test], data)
        result = dict(zip(meta_frame.columns, meta_frame.iloc[0]))
        expected = {**defaults, **test}
        assert result == expected


def test_variableRegex(data):

    tests = [
        (".*", ".*"),
        ("var(1|2)", "var(1|2)"),
        ("(.*3)", "(.*3)")
    ]

    for config_wc, expected_wc in tests:
        _, config = initMetaDict(
            [{F.VARNAME: config_wc, F.TESTS: "flagAll()"}],
            data
        )
        expected = data.columns[data.columns.str.match(expected_wc)]
        assert np.all(config[F.VARNAME] == expected)


def test_inlineComments(data):
    """
    adresses issue #3
    """
    config = f"""
    {F.VARNAME}|{F.TESTS}|{F.PLOT}
    pre2|flagAll() # test|False # test
    """
    _, meta_frame = initMetaString(config, data)
    assert meta_frame.loc[0, F.PLOT] == False
    assert meta_frame.loc[0, F.TESTS] == "flagAll()"


def test_configReaderLineNumbers(data):
    config = f"""
    {F.VARNAME}|{F.TESTS}
    #temp1|flagAll()
    pre1|flagAll()
    pre2|flagAll()
    SM|flagAll()
    #SM|flagAll()
    SM1|flagAll()
    """
    meta_fname, meta_frame = initMetaString(config, data)
    result = meta_frame[F.LINENUMBER].tolist()
    expected = [3, 4, 5, 7]
    assert result == expected


@pytest.mark.parametrize("flagger", TESTFLAGGER)
@pytest.mark.parametrize("nodata", TESTNODATA)
def test_configChecks(data, flagger, nodata, caplog):

    flagger = flagger.initFlags(data)
    flags = flagger.getFlags()
    var1, var2, var3, *_ = data.columns

    tests = [
        ({F.VARNAME: var1, F.TESTS: "range(mn=0)"}, TypeError),
        ({F.VARNAME: var3, F.TESTS: "flagNothing()"}, NameError),
        ({F.VARNAME: "", F.TESTS: "range(min=3)"}, SyntaxError),
        ({F.VARNAME: var1, F.TESTS: ""}, SyntaxError),
        ({F.VARNAME: ""}, SyntaxError),
        ({F.TESTS: "range(min=3)"}, SyntaxError),
    ]

    for config_dict, expected in tests:
        _, config_df = initMetaDict([config_dict], data)
        with pytest.raises(expected):
            checkConfig(config_df, data, flagger, nodata)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import io
import re

import numpy as np
import pandas as pd

from saqc.core.core import readConfig
from saqc.flagger import (
    ContinuousFlagger,
    CategoricalFlagger,
    SimpleFlagger,
    DmpFlagger,
)


TESTNODATA = (np.nan, -9999)


TESTFLAGGER = (
    CategoricalFlagger(["NIL", "GOOD", "BAD"]),
    SimpleFlagger(),
    DmpFlagger(),
    ContinuousFlagger(),
)


def initData(cols=2, start_date="2017-01-01", end_date="2017-12-31", freq="1h"):
    dates = pd.date_range(start=start_date, end=end_date, freq=freq)
    data = {}
    dummy = np.arange(len(dates))
    for col in range(1, cols + 1):
        data[f"var{col}"] = dummy * (col)
    return pd.DataFrame(data, index=dates)


def initMetaString(metastring, data):
    cleaned = re.sub(
        r"\s*,\s*", r",", re.sub(r"\|", r";", re.sub(r"\n[ \t]+", r"\n", metastring))
    )
    fobj = io.StringIO(cleaned.strip())
    config = readConfig(fobj, data)
    fobj.seek(0)
    return fobj, config


def _getKeys(metadict):
    keys = list(metadict[0].keys())
    for row in metadict[1:]:
        for k in row.keys():
            if k not in keys:
                keys.append(k)
    return keys


def initMetaDict(config_dict, data):
    df = pd.DataFrame(config_dict)[_getKeys(config_dict)]
    fobj = io.StringIO()
    df.fillna("").to_csv(fobj, index=False, sep=";")
    fobj.seek(0)
    config = readConfig(fobj, data)
    fobj.seek(0)
    return fobj, config



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import pytest
import numpy as np
import pandas as pd

from saqc.funcs.spike_detection import (
    flagSpikes_spektrumBased,
    flagSpikes_simpleMad,
    flagSpikes_slidingZscore,
    flagSpikes_basic,
)

from test.common import TESTFLAGGER


@pytest.fixture(scope="module")
def spiky_data():
    index = pd.date_range(start="2011-01-01", end="2011-01-05", freq="5min")
    spiky_series = pd.DataFrame(
        dict(spiky_data=np.linspace(1, 2, index.size)), index=index
    )
    spiky_series.iloc[100] = 100
    spiky_series.iloc[1000] = -100
    flag_assertion = [100, 1000]
    return spiky_series, flag_assertion


@pytest.mark.parametrize("flagger", TESTFLAGGER)
def test_flagSpikesSpektrumBased(spiky_data, flagger):
    data = spiky_data[0]
    field, *_ = data.columns
    flagger = flagger.initFlags(data)
    data, flagger_result = flagSpikes_spektrumBased(data, field, flagger)
    flag_result = flagger_result.getFlags(field)
    test_sum = (flag_result[spiky_data[1]] == flagger.BAD).sum()
    assert test_sum == len(spiky_data[1])


@pytest.mark.parametrize("flagger", TESTFLAGGER)
def test_flagMad(spiky_data, flagger):
    data = spiky_data[0]
    field, *_ = data.columns
    flagger = flagger.initFlags(data)
    data, flagger_result = flagSpikes_simpleMad(data, field, flagger, "1H")
    flag_result = flagger_result.getFlags(field)
    test_sum = (flag_result[spiky_data[1]] == flagger.BAD).sum()
    assert test_sum == len(spiky_data[1])


@pytest.mark.parametrize("flagger", TESTFLAGGER)
@pytest.mark.parametrize("method", ["modZ", "zscore"])
def test_slidingOutlier(spiky_data, flagger, method):

    # test for numeric input
    data = spiky_data[0]
    field, *_ = data.columns
    flagger = flagger.initFlags(data)

    tests = [
        flagSpikes_slidingZscore(data, field, flagger, window=300, offset=50, method=method),
        flagSpikes_slidingZscore(
            data, field, flagger, window="1500min", offset="250min", method=method
        ),
    ]

    for _, flagger_result in tests:
        flag_result = flagger_result.getFlags(field)
        test_sum = (flag_result.iloc[spiky_data[1]] == flagger.BAD).sum()
        assert int(test_sum) == len(spiky_data[1])


@pytest.mark.parametrize("flagger", TESTFLAGGER)
def test_flagSpikesBasic(spiky_data, flagger):
    data = spiky_data[0]
    field, *_ = data.columns
    flagger = flagger.initFlags(data)
    data, flagger_result = flagSpikes_basic(
        data, field, flagger, thresh=60, tolerance=10, window_size="20min"
    )
    flag_result = flagger_result.getFlags(field)
    test_sum = (flag_result[spiky_data[1]] == flagger.BAD).sum()
    assert test_sum == len(spiky_data[1])



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from collections import OrderedDict

import pandas as pd

from saqc.flagger.baseflagger import BaseFlagger
from saqc.lib.tools import assertDataFrame


class Flags(pd.CategoricalDtype):
    def __init__(self, flags):
        # NOTE: all flag schemes need to support
        #       at least 3 flag categories:
        #       * unflagged
        #       * good
        #       * bad
        assert len(flags) > 2
        super().__init__(flags, ordered=True)


class CategoricalFlagger(BaseFlagger):
    def __init__(self, flags):
        super().__init__(dtype=Flags(flags))
        self._categories = self.dtype.categories

    def _isDtype(self, flag) -> bool:
        """
        not needed here, move out
        """
        if isinstance(flag, pd.Series):
            return isinstance(flag.dtype, pd.CategoricalDtype) and flag.dtype == self.dtype
        return flag in self.dtype.categories

    @property
    def UNFLAGGED(self):
        return self._categories[0]

    @property
    def GOOD(self):
        return self._categories[1]

    @property
    def BAD(self):
        return self._categories[-1]

    def isSUSPICIOUS(self, flag):
        return flag in self._categories[2:-1]



#! /usr/bin/env python
# -*- coding: utf-8 -*-
import subprocess
import json
from copy import deepcopy
from collections import OrderedDict
from typing import Union, Sequence

import pandas as pd

from saqc.flagger.categoricalflagger import CategoricalFlagger
from saqc.lib.tools import assertDataFrame, toSequence, assertScalar


class Keywords:
    VERSION = "$version"


class FlagFields:
    FLAG = "quality_flag"
    CAUSE = "quality_cause"
    COMMENT = "quality_comment"


class ColumnLevels:
    VARIABLES = "variables"
    FLAGS = "flags"


FLAGS = ["NIL", "OK", "DOUBTFUL", "BAD"]


class DmpFlagger(CategoricalFlagger):
    def __init__(self):
        super().__init__(FLAGS)
        self.flags_fields = [FlagFields.FLAG, FlagFields.CAUSE, FlagFields.COMMENT]
        version = subprocess.run(
            "git describe --tags --always --dirty", shell=True, check=False, stdout=subprocess.PIPE,
        ).stdout
        self.project_version = version.decode().strip()
        self.signature = ("flag", "comment", "cause", "force")
        self._flags = None

    def initFlags(self, data: pd.DataFrame = None, flags: pd.DataFrame = None):
        """
        initialize a flagger based on the given 'data' or 'flags'
        if 'data' is not None: return a flagger with flagger.UNFALGGED values
        if 'flags' is not None: return a flagger with the given flags
        """

        if data is not None:
            flags = pd.DataFrame(data=self.UNFLAGGED, columns=self._getColumnIndex(data.columns), index=data.index,)
        elif flags is not None:
            if not isinstance(flags.columns, pd.MultiIndex):
                flags = flags.T.set_index(keys=self._getColumnIndex(flags.columns, [FlagFields.FLAG])).T.reindex(
                    columns=self._getColumnIndex(flags.columns)
                )
        else:
            raise TypeError("either 'data' or 'flags' are required")

        return self._copy(self._assureDtype(flags))

    def getFlagger(self, field=None, loc=None, iloc=None):
        # NOTE: we need to preserve all indexing levels
        assertScalar("field", field, optional=True)
        variables = self._flags.columns.get_level_values(ColumnLevels.VARIABLES).drop_duplicates()
        cols = toSequence(field, variables)
        out = super().getFlagger(field, loc, iloc)
        out._flags.columns = self._getColumnIndex(cols)
        return out

    def getFlags(self, field=None, loc=None, iloc=None):
        assertScalar("field", field, optional=True)
        field = field or slice(None)
        mask = self._locatorMask(field, loc, iloc)
        flags = self._flags.xs(FlagFields.FLAG, level=ColumnLevels.FLAGS, axis=1).copy()
        return super()._assureDtype(flags.loc[mask, field])

    def setFlags(self, field, loc=None, iloc=None, flag=None, force=False, comment="", cause="", **kwargs):
        assertScalar("field", field, optional=True)

        flag = self.BAD if flag is None else self._checkFlag(flag)

        comment = json.dumps({"comment": comment, "commit": self.project_version, "test": kwargs.get("func_name", ""),})

        this = self.getFlags(field=field)
        other = self._broadcastFlags(field=field, flag=flag)
        mask = self._locatorMask(field, loc, iloc)
        if not force:
            mask &= (this < other).values

        out = deepcopy(self)
        out._flags.loc[mask, field] = other[mask], cause, comment
        return out

    def _getColumnIndex(
        self, cols: Union[str, Sequence[str]], fields: Union[str, Sequence[str]] = None
    ) -> pd.MultiIndex:
        cols = toSequence(cols)
        fields = toSequence(fields, self.flags_fields)
        return pd.MultiIndex.from_product([cols, fields], names=[ColumnLevels.VARIABLES, ColumnLevels.FLAGS])

    def _assureDtype(self, flags):
        # NOTE: building up new DataFrames is significantly
        #       faster than assigning into existing ones
        tmp = OrderedDict()
        for (var, flag_field) in flags.columns:
            col_data = flags[(var, flag_field)]
            if flag_field == FlagFields.FLAG:
                col_data = col_data.astype(self.dtype)
            else:
                col_data = col_data.astype(str)
            tmp[(var, flag_field)] = col_data
        return pd.DataFrame(tmp, columns=flags.columns, index=flags.index)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import numbers
from typing import Sequence, Union, Any, Iterator

import numpy as np
import pandas as pd
import numba as nb

# from saqc.flagger import BaseFlagger
from saqc.lib.types import T, PandasLike

STRING_2_FUNC = {
    "sum": np.sum,
    "mean": np.mean,
    "median": np.median,
    "min": np.min,
    "max": np.max,
    "first": pd.Series(np.nan, index=pd.DatetimeIndex([])).resample("0min").first,
    "last": pd.Series(np.nan, index=pd.DatetimeIndex([])).resample("0min").last,
}


def assertScalar(name, value, optional=False):
    if (not np.isscalar(value)) and (value is not None) and (optional is True):
        raise ValueError(f"'{name}' needs to be a scalar or 'None'")
    elif (not np.isscalar(value)) and optional is False:
        raise ValueError(f"'{name}' needs to be a scalar")


def toSequence(value: Union[T, Sequence[T]], default: Union[T, Sequence[T]] = None) -> Sequence[T]:
    if value is None:
        value = default
    if np.isscalar(value):
        value = [value]
    return value


@nb.jit(nopython=True, cache=True)
def findIndex(iterable, value, start):
    i = start
    while i < len(iterable):
        v = iterable[i]
        if v >= value:
            return i
        i = i + 1
    return -1


@nb.jit(nopython=True, cache=True)
def valueRange(iterable):
    minval = iterable[0]
    maxval = minval
    for v in iterable:
        if v < minval:
            minval = v
        elif v > maxval:
            maxval = v
    return maxval - minval


def slidingWindowIndices(dates, window_size, iter_delta=None):
    """
    this function is a building block of a custom implementation of
    the pandas 'rolling' method. A number of shortcomings in the
    'rolling' implementation might made this a worthwhil endavour,
    namly:
    + There is no way to provide a step size, i.e. to not start the
      next rolling window at the very next row in the DataFrame/Series
    + The inconsistent bahaviour with numerical vs frequency based
      window sizes. When winsz is an integer, all windows are equally
      large (winsz=5 -> windows contain 5 elements), but variable in
      size, when the winsz is a frequency string (winsz="2D" ->
      window grows from size 1 during the first iteration until it
      covers the given frequency). Especially the bahaviour with
      frequency strings is quite unfortunate when calling methods
      relying on the size of the window (sum, mean, median)
    """

    # lets work on numpy data structures for performance reasons
    if isinstance(dates, (pd.DataFrame, pd.Series)):
        dates = dates.index
    dates = np.array(dates, dtype=np.int64)

    if np.any(np.diff(dates) <= 0):
        raise ValueError("strictly monotonic index needed")

    window_size = pd.to_timedelta(window_size).to_timedelta64().astype(np.int64)
    if iter_delta:
        iter_delta = pd.to_timedelta(iter_delta).to_timedelta64().astype(np.int64)

    start_date = dates[0]
    last_date = dates[-1]
    start_idx = 0
    end_idx = start_idx

    while True:
        end_date = start_date + window_size
        if (end_date > last_date) or (start_idx == -1) or (end_idx == -1):
            break

        end_idx = findIndex(dates, end_date, end_idx)
        yield start_idx, end_idx

        if iter_delta:
            start_idx = findIndex(dates, start_date + iter_delta, start_idx)
        else:
            start_idx += 1

        start_date = dates[start_idx]


def inferFrequency(data: PandasLike) -> pd.DateOffset:
    return pd.tseries.frequencies.to_offset(pd.infer_freq(data.index))


def combineDataFrames(left: pd.DataFrame, right: pd.DataFrame, fill_value: float = np.nan) -> pd.DataFrame:
    """
    Combine the given DataFrames 'left' and 'right' such that, the
    output is union of the indices and the columns of both. In case
    of duplicated values, 'left' is overwritten by 'right'
    """
    combined = left.reindex(
        index=left.index.union(right.index),
        columns=left.columns.union(right.columns, sort=False),
        fill_value=fill_value,
    )

    for key, values in right.iteritems():
        combined.loc[right.index, key] = values

    return combined


def retrieveTrustworthyOriginal(data: pd.DataFrame, field: str, flagger=None, level: Any = None) -> pd.DataFrame:
    """Columns of data passed to the saqc runner may not be sampled to its original sampling rate - thus
    differenciating between missng value - nans und fillvalue nans is impossible.

    This function:
    (1) if flagger is None:
        (a) estimates the sampling rate of the input dataseries by dropping all nans and then returns the series at the
            estimated samplng rate.

    (2) if "flagger" is not None but "level" is None:
        (a) all values are dropped, that are flagged worse then flagger.GOOD. (so unflagged values wont be dropped)
        (b) estimates the sampling rate of the input dataseries by dropping all nans and then returns the series at the
            estimated samplng rate.
    (3) if "flagger" is not None and "level" is not None:
        (a) all values are dropped, that are flagged worse then level. (so unflagged values wont be dropped)
        (b) estimates the sampling rate of the input dataseries by dropping all nans and then returns the series at the
            estimated samplng rate.

    Note, that the passed dataseries should be harmonized to an equidistant
        frequencie grid (maybe including blow up entries).

    :param data:        DataFrame. The Data frame holding the data containing 'field'.
    :param field:       String. Fieldname of the column in data, that you want to sample to original sampling rate.
                        It has to have a harmonic
    :param flagger:     None or a flagger object.
    :param level:       Lower bound of flags that are excepted for data. Must be a flag the flagger can handle.

    """
    dataseries = data[field]

    if flagger is not None:
        mask = flagger.isFlagged(field, flag=level or flagger.GOOD, comparator="<=")
        # drop all flags that are suspicious or worse
        dataseries = dataseries[mask]

    # drop the nan values that may result from any preceeding upsampling of the measurements:
    dataseries = dataseries.dropna()

    if dataseries.empty:
        return dataseries, np.nan

    # estimate original data sampling frequencie
    # (the original series sampling rate may not match data-input sample rate):
    seconds_rate = dataseries.index.to_series().diff().min().seconds
    data_rate = pd.tseries.frequencies.to_offset(str(seconds_rate) + "s")

    return dataseries.asfreq(data_rate), data_rate


def offset2seconds(offset):
    """Function returns total seconds upon "offset like input

    :param offset:  offset string or pandas offset object.
    """

    return pd.Timedelta.total_seconds(pd.Timedelta(offset))


def flagWindow(flagger_old, flagger_new, field, direction="fw", window=0, **kwargs) -> pd.Series:

    if window == 0 or window == "":
        return flagger_new

    fw, bw = False, False
    mask = flagger_old.getFlags(field) != flagger_new.getFlags(field)
    f = flagger_new.isFlagged(field) & mask

    if not mask.any():
        # nothing was flagged, so nothing need to be flagged additional
        return flagger_new

    if isinstance(window, int):
        x = f.rolling(window=window + 1).sum()
        if direction in ["fw", "both"]:
            fw = x.fillna(method="bfill").astype(bool)
        if direction in ["bw", "both"]:
            bw = x.shift(-window).fillna(method="bfill").astype(bool)
    else:
        # time-based windows
        if direction in ["bw", "both"]:
            # todo: implement time-based backward rolling
            raise NotImplementedError
        fw = f.rolling(window=window, closed="both").sum().astype(bool)

    fmask = bw | fw
    return flagger_new.setFlags(field, fmask, **kwargs)


def sesonalMask(dtindex, month0=1, day0=1, month1=12, day1=None):
    """
    This function provide a mask for a sesonal time range in the given dtindex.
    This means the interval is applied again on every year and even over the change of a year.
    Note that both edges are inclusive.

    Examples:
        sesonal(dtindex, 1, 1, 3, 1)  -> [jan-mar]
        sesonal(dtindex, 8, 1, 8, 15) -> [1.aug-15.aug]


    This also works, if the second border is smaller then the first

    Examples:
        sesonal(dtindex, 10, 1, 2, 1) -> [1.nov-1.feb (following year)]
        sesonal(dtindex, 1, 10, 1, 1)  -> [10.jan-1.jan(following year)] like everything except ]1.jan-10.jan[

    """
    if day1 is None:
        day1 = 31 if month1 in [1, 3, 5, 7, 8, 10, 12] else 29 if month1 == 2 else 30

    # test plausibility of date
    try:
        f = "%Y-%m-%d"
        t0 = pd.to_datetime(f"2001-{month0}-{day0}", format=f)
        t1 = pd.to_datetime(f"2001-{month1}-{day1}", format=f)
    except ValueError:
        raise ValueError("Given datelike parameter not logical")

    # swap
    if t1 < t0:
        # we create the same mask as we would do if not inverted
        # but the borders need special treatment..
        # ===end]....................[start====
        # ======]end+1........start-1[=========
        # ......[end+1========start-1]......... + invert
        # ......[start`========= end`]......... + invert
        t0 -= pd.to_timedelta("1d")
        t1 += pd.to_timedelta("1d")
        invert = True
        # only swap id condition is still true
        t0, t1 = t1, t0 if t1 < t0 else (t0, t1)

        month0, day0 = t0.month, t0.day
        month1, day1 = t1.month, t1.day
    else:
        invert = False

    month = [m for m in range(month0, month1 + 1)]

    # make a mask for [start:end]
    mask = dtindex.month.isin(month)
    if day0 > 1:
        exclude = [d for d in range(1, day0)]
        mask &= ~(dtindex.month.isin([month0]) & dtindex.day.isin(exclude))
    if day1 < 31:
        exclude = [d for d in range(day1 + 1, 31 + 1)]
        mask &= ~(dtindex.month.isin([month1]) & dtindex.day.isin(exclude))

    if invert:
        return ~mask
    else:
        return mask


def assertDataFrame(df: Any, argname: str = "arg", allow_multiindex: bool = True) -> None:
    if not isinstance(df, pd.DataFrame):
        raise TypeError(f"{argname} must be of type pd.DataFrame, {type(df)} was given")
    if not allow_multiindex:
        assertSingleColumns(df, argname)
    if not df.columns.is_unique:
        raise TypeError(f"{argname} must have unique columns")


def assertSeries(srs: Any, argname: str = "arg") -> None:
    if not isinstance(srs, pd.Series):
        raise TypeError(f"{argname} must be of type pd.Series, {type(srs)} was given")


def assertPandas(pdlike: PandasLike, argname: str = "arg", allow_multiindex: bool = True) -> None:
    if not isinstance(pdlike, pd.Series) and not isinstance(pdlike, pd.DataFrame):
        raise TypeError(f"{argname} must be of type pd.DataFrame or pd.Series, {type(pdlike)} was given")
    if not allow_multiindex:
        assertSingleColumns(pdlike, argname)


def assertMultiColumns(dfmi: pd.DataFrame, argname: str = "") -> None:
    assertDataFrame(dfmi, argname, allow_multiindex=True)
    if not isinstance(dfmi.columns, pd.MultiIndex):
        raise TypeError(
            f"given pd.DataFrame ({argname}) need to have a muliindex on columns, "
            f"instead it has a {type(dfmi.columns)}"
        )


def assertSingleColumns(df: PandasLike, argname: str = "") -> None:
    if isinstance(df, pd.DataFrame) and isinstance(df.columns, pd.MultiIndex):
        raise TypeError(f"given pd.DataFrame {argname} is not allowed to have a muliindex on columns")


def getFuncFromInput(func):
    """
    Aggregation functions passed by the user, are selected by looking them up in the STRING_2_DICT dictionary -
    But since there are wrappers, that dynamically generate aggregation functions and pass those on ,the parameter
    interfaces must as well be capable of processing real functions passed. This function does that.

    :param func: A key to the STRING_2_FUNC dict, or an actual function
    """
    # if input is a callable - than just pass it:
    if hasattr(func, "__call__"):
        if (func.__name__ == "aggregator") & (func.__module__ == "saqc.funcs.harm_functions"):
            return func
        else:
            raise ValueError("The function you passed is suspicious!")
    elif func in STRING_2_FUNC.keys():
        return STRING_2_FUNC[func]
    else:
        raise ValueError("Function input not a callable nor a known key to internal the func dictionary.")


@nb.jit(nopython=True, cache=True)
def otherIndex(values: np.ndarray, start: int = 0) -> int:
    """
    returns the index of the first non value not equal to values[0]
    -> values[start:i] are all identical
    """
    val = values[start]
    for i in range(start, len(values)):
        if values[i] != val:
            return i
    return -1


def groupConsecutives(series: pd.Series) -> Iterator[pd.Series]:

    """
    group consecutive values into distinct pd.Series
    """
    index = series.index
    values = series.values
    target = values[0]

    start = 0
    while True:
        stop = otherIndex(values, start)
        if stop == -1:
            break
        yield pd.Series(data=values[start:stop], index=index[start:stop])
        start = stop



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from typing import TypeVar

import numpy as np
import pandas as pd

T = TypeVar("T")
ArrayLike = TypeVar("ArrayLike", np.ndarray, pd.Series, pd.DataFrame)
PandasLike = TypeVar("PandasLike", pd.Series, pd.DataFrame)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import pandas as pd

from saqc.funcs.register import register
from saqc.lib.statistic_functions import varQC
from saqc.lib.tools import (
    valueRange,
    slidingWindowIndices,
    retrieveTrustworthyOriginal,
    groupConsecutives,
)


@register("constant")
def flagConstant(data, field, flagger, thresh, window, **kwargs):
    """
    Flag values are (semi-)constant.

    :param data: dataframe
    :param field: column in data
    :param flagger: saqc flagger obj
    :param thresh: the difference between two values must be below that
    :param window: sliding window
    """
    d = data[field]

    # find all constant values in a row with a forward search
    r = d.rolling(window=window)
    mask = (r.max() - r.min() <= thresh) & (r.count() > 1)

    # backward rolling for offset windows hack
    bw = mask[::-1].copy()
    bw.index = bw.index.max() - bw.index

    # propagate the mask(!), backwards
    bwmask = bw.rolling(window=window).sum() > 0

    mask |= bwmask[::-1].values

    flagger = flagger.setFlags(field, mask, **kwargs)
    return data, flagger


@register("constant_varianceBased")
def flagConstantVarianceBased(
    data, field, flagger, window="12h", thresh=0.0005, max_missing=None, max_consec_missing=None, **kwargs
):

    """
    Function flags plateaus/series of constant values. Any interval of values y(t),..y(t+n) is flagged, if:

    (1) n > "plateau_interval_min"
    (2) variance(y(t),...,y(t+n) < thresh

    :param data:                        The pandas dataframe holding the data-to-be flagged.
                                        Data must be indexed by a datetime series and be harmonized onto a
                                        time raster with seconds precision (skips allowed).
    :param field:                       Fieldname of the Soil moisture measurements field in data.
    :param flagger:                     A flagger - object. (saqc.flagger.X)
    :param window:                      Offset String. Only intervals of minimum size "window" have the
                                        chance to get flagged as constant intervals
    :param thresh:                      Float. The upper barrier, the variance of an interval mus not exceed, if the
                                        interval wants to be flagged a plateau.
    :param max_missing:                 maximum number of nan values tolerated in an interval, for retrieving a valid
                                        variance from it. (Intervals with a number of nans exceeding "max_missing"
                                        have no chance to get flagged a plateau!)
    :param max_consec_missing:          Maximum number of consecutive nan values allowed in an interval to retrieve a
                                        valid  variance from it. (Intervals with a number of nans exceeding
                                        "max_missing" have no chance to get flagged a plateau!)
    """

    dataseries, data_rate = retrieveTrustworthyOriginal(data, field, flagger)

    if max_missing is None:
        max_missing = np.inf
    if max_consec_missing is None:
        max_consec_missing = np.inf

    min_periods = int(np.ceil(pd.Timedelta(window) / pd.Timedelta(data_rate)))

    plateaus = dataseries.rolling(window=window, min_periods=min_periods).apply(
        lambda x: True if varQC(x, max_missing, max_consec_missing) <= thresh else np.nan, raw=False,
    )

    # are there any candidates for beeing flagged plateau-ish
    if plateaus.sum() == 0:
        return data, flagger

    plateaus.fillna(method="bfill", limit=min_periods - 1, inplace=True)

    # result:
    plateaus = (plateaus[plateaus == 1.0]).index

    flagger = flagger.setFlags(field, plateaus, **kwargs)
    return data, flagger



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import pandas as pd
import numpy as np
import logging

from saqc.funcs.functions import flagMissing
from saqc.funcs.register import register
from saqc.lib.tools import toSequence, getFuncFromInput


class Heap:
    INDEX = "initial_ts"
    DATA = "original_data"
    FLAGGER = "original_flagger"
    FREQ = "freq"
    METHOD = "reshape_method"
    DROP = "drop_flags"


HARM_2_DEHARM = {
    "fshift": "invert_fshift",
    "bshift": "invert_bshift",
    "nshift": "invert_nearest",
    "fagg": "invert_fshift",
    "bagg": "invert_bshift",
    "nagg": "invert_nearest",
    "fagg_no_deharm": "regain",
    "bagg_no_deharm": "regain",
    "nagg_no_deharm": "regain",
}


def harmWrapper(heap={}):
    # NOTE:
    # (1) - harmonization will ALWAYS flag flagger.BAD all the np.nan values and afterwards DROP ALL
    #       flagger.BAD flagged values from flags frame for further flagging!!!!!!!!!!!!!!!!!!!!!
    def harmonize(
        data,
        field,
        flagger,
        freq,
        inter_method,
        reshape_method,
        inter_agg="mean",
        inter_order=1,
        inter_downcast=False,
        reshape_agg="max",
        reshape_missing_flag=None,
        reshape_shift_comment=False,
        drop_flags=None,
        data_missing_value=np.nan,
        **kwargs,
    ):

        # get funcs from strings:
        inter_agg = getFuncFromInput(inter_agg)
        reshape_agg = getFuncFromInput(reshape_agg)

        # for some tingle tangle reasons, resolving the harmonization will not be sound, if not all missing/np.nan
        # values get flagged initially:
        data, flagger = flagMissing(data, field, flagger, nodata=data_missing_value, **kwargs)
        # and dropped for harmonization:
        if drop_flags is not None:
            if flagger.BAD not in drop_flags:
                drop_flags.append(flagger.BAD)

        # before sending the current flags and data frame to the future (for backtracking reasons), we clear it
        # from merge-nans that just resulted from harmonization of other variables!
        dat_col, flagger_merged = _fromMerged(data, flagger, field)

        # now we send the flags frame in its current shape to the future:
        heap[field] = {
            Heap.DATA: dat_col,
            Heap.FLAGGER: flagger_merged,
            Heap.FREQ: freq,
            Heap.METHOD: reshape_method,
            Heap.DROP: drop_flags,
        }

        # furthermore we need to memorize the initial timestamp to ensure output format will equal input format.
        if Heap.INDEX not in heap.keys():
            heap.update({Heap.INDEX: dat_col.index})

        # now we can manipulate it without loosing information gathered before harmonization
        dat_col, flagger_merged_clean = _outsortCrap(dat_col, field, flagger_merged, drop_flags=drop_flags,)

        # interpolation! (yeah)
        dat_col, chunk_bounds = _interpolateGrid(
            dat_col,
            freq,
            method=inter_method,
            order=inter_order,
            agg_method=inter_agg,
            total_range=(heap[Heap.INDEX][0], heap[Heap.INDEX][-1]),
            downcast_interpolation=inter_downcast,
        )

        # flags now have to be carefully adjusted according to the changes/shifts we did to data
        flagger_merged_clean_reshaped = _reshapeFlags(
            flagger_merged_clean,
            field,
            ref_index=dat_col.index,
            method=reshape_method,
            agg_method=reshape_agg,
            missing_flag=reshape_missing_flag,
            set_shift_comment=reshape_shift_comment,
            block_flags=chunk_bounds,
            **kwargs,
        )

        # finally we happily blow up the data and flags frame again,
        # to release them on their ongoing journey through saqc.
        data, flagger_out = _toMerged(
            data, flagger, field, data_to_insert=dat_col, flagger_to_insert=flagger_merged_clean_reshaped,
        )

        return data, flagger_out

    def deharmonize(data, field, flagger, co_flagging=False, **kwargs):

        # Check if there is backtracking information available for actual harmonization resolving
        if field not in heap:
            logging.warning(
                'No backtracking data for resolving harmonization of "{}". Reverse projection of flags gets'
                " skipped!".format(field)
            )
            return data, flagger

        # get some deharm configuration infos from the heap:
        harm_info = heap.pop(field)
        resolve_method = HARM_2_DEHARM[harm_info[Heap.METHOD]]

        # retrieve data and flags from the merged saqc-conform data frame (and by that get rid of blow-up entries).
        dat_col, flagger_merged = _fromMerged(data, flagger, field)

        # reconstruct the drops that were performed before harmonization
        drops, flagger_original_clean = _outsortCrap(
            dat_col, field, harm_info[Heap.FLAGGER], drop_flags=harm_info[Heap.DROP], return_drops=True,
        )

        # with reconstructed pre-harmonization flags-frame -> perform the projection of the flags calculated for
        # the harmonized timeseries, onto the original timestamps
        flagger_back = _backtrackFlags(
            flagger_merged,
            flagger_original_clean,
            harm_info[Heap.FREQ],
            track_method=resolve_method,
            co_flagging=co_flagging,
        )
        flags_back = flagger_back.getFlags()

        # now: re-insert the pre-harmonization-drops
        flags_col = flags_back.reindex(flags_back.index.join(drops.index, how="outer"))
        # due to assignment reluctants with 1-d-dataframes we are squeezing:
        flags_col = flags_col.squeeze(axis=1)
        drops = drops.squeeze(axis=1)
        flags_col.loc[drops.index] = drops

        # but to stick with the policy of always having flags as pd.DataFrames we blow up the flags col again:
        if isinstance(flags_col, pd.Series):
            flags_col = flags_col.to_frame()
        flagger_back_full = flagger.initFlags(flags=flags_col)

        dat_col = harm_info[Heap.DATA].reindex(flags_col.index, fill_value=np.nan)
        dat_col.name = field
        # transform the result into the form, data travels through saqc:
        data, flagger_out = _toMerged(data, flagger, field, dat_col, flagger_back_full, target_index=heap[Heap.INDEX],)
        # clear heap if nessecary:
        if len(heap) == 1 and Heap.INDEX in heap:
            del heap[Heap.INDEX]

        # bye bye data
        return data, flagger_out

    return harmonize, deharmonize


harmonize, deharmonize = harmWrapper()
register("harmonize")(harmonize)
register("deharmonize")(deharmonize)


# (de-)harmonize helper
def _outsortCrap(
    data, field, flagger, drop_flags=None, return_drops=False,
):

    """Harmonization gets the more easy, the more data points we can exclude from crowded sampling intervals.
    Depending on passed key word options the function will remove nan entries and as-suspicious-flagged values from
    the data and the flags passed. In deharmonization the function is used to reconstruct original flags field shape.

    :param data:            pd.Series. ['data'].
    :param flagger:         saqc.flagger.
    :param drop_suspicious: Boolean. Default = True. If True, only values that are flagged GOOD or UNFLAGGED get
                            processed.
    :param drop_bad:        Boolean. Default = True. If True, BAD-flagged values get dropped from data.
    :param drop_list:       List or None. Default = None. List of flags that shall be dropped from data. If None is
                            passed (default), list based data dropping is omitted.
    :param return_drops:    Boolean. Default = False. If True, return the drops only. If False, return the data and
                            flags without drops.
    :return:                If return_drops=False. (default) Returns data, flags tuple with values-to-be-dropped
                            dropped.
                            If return_drops=True. Returns the dropped flags.
    """

    drop_mask = pd.Series(data=False, index=data.index)

    drop_flags = toSequence(drop_flags, default=flagger.BAD)
    for drop_flag in drop_flags:
        drop_mask = drop_mask | flagger.isFlagged(field, flag=drop_flag, comparator="==")

    flagger_out = flagger.getFlagger(loc=~drop_mask)
    if return_drops:
        return flagger.getFlags(loc=drop_mask), flagger_out
    return data[~drop_mask], flagger_out


def _makeGrid(t0, t1, freq, name=None):
    """
    Returns a frequency grid, covering the date range of 'data'.
    :param data:    pd.Series. ['data']
    :param freq:    Offset String. Intended Sampling frequency.
    :return:        pd.Series. ['data'].
    """

    harm_start = t0.floor(freq=freq)
    harm_end = t1.ceil(freq=freq)
    return pd.date_range(start=harm_start, end=harm_end, freq=freq, name=name)


def _insertGrid(data, freq):
    """
    Depending on the frequency, the data has to be harmonized to, the passed data series gets reindexed with an index,
    containing the 'original' entries and additionally, if not present, the equidistant entries of the frequency grid.
    :param data:    pd.Series. ['data']
    :param freq:    Offset String. Intended Sampling frequency.
    :return:        pd.Series. ['data'].
    """

    return data.reindex(
        data.index.join(_makeGrid(data.index[0], data.index[-1], freq, name=data.index.name), how="outer",)
    )


def _interpolateGrid(
    data, freq, method, order=1, agg_method=sum, total_range=None, downcast_interpolation=False,
):
    """The function calculates grid point values for a passed pd.Series (['data']) by applying
    the selected interpolation/fill method. (passed to key word 'method'). The interpolation will apply for grid points
    only, that have preceding (forward-aggregation/forward-shifts) or succeeding (backward-aggregation/backward-shift)
    values, or both ("real" interpolations, like linear, polynomial, ...-> see documentation below).

    Data must be cleared from nans before entering here.

    Methods:
    All Methods calculate new values at grid points, if there is no data available for that sampling point.

    1. "real" INTERPOLATIONS

    There are available all the interpolation methods from the pandas.interpolate() method and they are applicable by
    the very same key words, that you would pass to pd.Series.interpolates's method parameter.

    Be careful with pd.Series.interpolate's 'nearest' and 'pad':
    To just fill grid points forward/backward or from the nearest point - and
    assign grid points, that refer to missing data, a nan value, the use of "fshift", "bshift" and "nshift" is
    recommended, to ensure the result expected. (The methods diverge in some special cases).

    To make an index-aware linear interpolation, "times" has to be passed - NOT 'linear'.

    2. SHIFTS:

    'fshift'        -  every grid point gets assigned its ultimately preceeding value - if there is one available in
                    the preceeding sampling interval.
    'bshift'        -  every grid point gets assigned its first succeeding value - if there is one available in the
                    succeeding sampling interval.
    'nshift' -  every grid point gets assigned the nearest value in its range ( range = +/-(freq/2) ).

    3. AGGREGATIONS

    'nagg'   - all values in the range (+/- freq/2) of a grid point get aggregated with agg_method and assigned
                    to it.
    'bagg'          - all values in a sampling interval get aggregated with agg_method and the result gets assigned to
                    the last grid point
    'fagg'          - all values in a sampling interval get aggregated with agg_method and the result gets assigned to
                    the next grid point

    :param data:        pd.DataFrame. ['data'].
    :param freq:        Offset String. the grid frequency.
    :param method:      String. Method you want to interpolate with. See function doc above.
    :param order:       Integer. Default = 1. If an interpolation method is selected that needs
                        to know about its "order", this is where you pass it. (For example 'polynomial', 'spline')
    :param agg_method:  Func. Default = sum. If an aggregation method is selected for grid point filling,
                        you need to pass the aggregation method to this very parameter. Note that it should be able
                        to handle empty argument series passed as well as np.nan passed.
    :param total_range  2-Tuple of pandas Timestamps.
                        The total range of all the data in the Dataframe that is currently processed. If not
                        None, the resulting harmonization grid of the current data column will range over the total
                        Data-range. This ensures not having nan-entries in the flags dataframe after harmonization.
    :return:            pd.DataFrame. ['data'].
    """

    chunk_bounds = None
    aggregations = ["nagg", "bagg", "fagg"]
    shifts = ["fshift", "bshift", "nshift"]
    interpolations = [
        "linear",
        "time",
        "nearest",
        "zero",
        "slinear",
        "quadratic",
        "cubic",
        "spline",
        "barycentric",
        "polynomial",
        "krogh",
        "piecewise_polynomial",
        "spline",
        "pchip",
        "akima",
    ]
    data = data.copy()
    ref_index = _makeGrid(data.index[0], data.index[-1], freq, name=data.index.name)
    if total_range is not None:
        total_index = _makeGrid(total_range[0], total_range[1], freq, name=data.index.name)

    # Aggregations:
    if method in aggregations:
        if method == "nagg":
            # all values within a grid points range (+/- freq/2, closed to the left) get aggregated with 'agg method'
            # some timestamp acrobatics to feed the base keyword properly
            seconds_total = pd.Timedelta(freq).total_seconds()
            seconds_string = str(int(seconds_total)) + "s"
            # calculate the series of aggregated values
            data = data.resample(seconds_string, base=seconds_total / 2, loffset=pd.Timedelta(freq) / 2).apply(
                agg_method
            )

        elif method == "bagg":
            # all values in a sampling interval get aggregated with agg_method and assigned to the last grid point
            data = data.resample(freq).apply(agg_method)
        # if method is fagg
        else:
            # all values in a sampling interval get aggregated with agg_method and assigned to the next grid point
            data = data.resample(freq, closed="right", label="right").apply(agg_method)
        # some consistency cleanup:
        if total_range is None:
            data = data.reindex(ref_index)

    # Shifts
    elif method in shifts:
        if method == "fshift":
            direction = "ffill"
            tolerance = pd.Timedelta(freq)

        elif method == "bshift":
            direction = "bfill"
            tolerance = pd.Timedelta(freq)
        # if method = nshift
        else:
            direction = "nearest"
            tolerance = pd.Timedelta(freq) / 2

        data = data.reindex(ref_index, method=direction, tolerance=tolerance)

    # Interpolations:
    elif method in interpolations:

        data = _insertGrid(data, freq)
        data, chunk_bounds = _interpolate(
            data, method, order=order, inter_limit=2, downcast_interpolation=downcast_interpolation,
        )
        if total_range is None:
            data = data.asfreq(freq, fill_value=np.nan)

    else:
        methods = "\n".join([", ".join(shifts), ", ".join(aggregations), ", ".join(interpolations)])
        raise ValueError(f"Unknown interpolation method: '{method}', please select from:\n{methods}")

    if total_range is not None:
        data = data.reindex(total_index)

    return data, chunk_bounds


def _interpolate(data, method, order=2, inter_limit=2, downcast_interpolation=False):
    """
    The function interpolates nan-values (and nan-grids) in timeseries data. It can be passed all the method keywords
    from the pd.Series.interpolate method and will than apply this very methods. Note, that the inter_limit keyword
    really restricts the interpolation to chunks, not containing more than "inter_limit" nan entries
    (thereby opposing the limit keyword of pd.Series.interpolate).

    :param data:                    pd.Series. The data series to be interpolated
    :param method:                  String. Method keyword designating interpolation method to use.
    :param order:                   Integer. If your desired interpolation method needs an order to be passed -
                                    here you pass it.
    :param inter_limit:             Integer. Default = 2. Limit up to wich nan - gaps in the data get interpolated.
                                    Its default value suits an interpolation that only will apply on an inserted
                                    frequency grid.
    :param downcast_interpolation:  Boolean. Default False. If True:
                                    If a data chunk not contains enough values for interpolation of the order "order",
                                    the highest order possible will be selected for that chunks interpolation."
    :return:
    """

    gap_mask = (data.rolling(inter_limit, min_periods=0).apply(lambda x: np.sum(np.isnan(x)), raw=True)) != inter_limit

    if inter_limit == 2:
        gap_mask = gap_mask & gap_mask.shift(-1, fill_value=True)
    else:
        gap_mask = (
            gap_mask.replace(True, np.nan).fillna(method="bfill", limit=inter_limit).replace(np.nan, True).astype(bool)
        )
    # start end ending points of interpolation chunks have to be memorized to block their flagging:
    chunk_switches = gap_mask.astype(int).diff()
    chunk_starts = chunk_switches[chunk_switches == -1].index
    chunk_ends = chunk_switches[(chunk_switches.shift(-1) == 1)].index
    chunk_bounds = chunk_starts.join(chunk_ends, how="outer", sort=True)

    data = data[gap_mask]

    if method in ["linear", "time"]:

        data.interpolate(method=method, inplace=True, limit=1, limit_area="inside")

    else:
        dat_name = data.name
        gap_mask = (~gap_mask).cumsum()
        data = pd.merge(gap_mask, data, how="inner", left_index=True, right_index=True)

        def _interpolWrapper(x, wrap_order=order, wrap_method=method):
            if x.count() > wrap_order:
                try:
                    return x.interpolate(method=wrap_method, order=int(wrap_order))
                except (NotImplementedError, ValueError):
                    logging.warning(
                        "Interpolation with method {} is not supported at order {}. "
                        "Interpolation will be performed with order {}".format(
                            method, str(wrap_order), str(wrap_order - 1)
                        )
                    )
                    return _interpolWrapper(x, int(wrap_order - 1), wrap_method)
            elif x.size < 3:
                return x
            else:
                if downcast_interpolation:
                    return _interpolWrapper(x, int(x.count() - 1), wrap_method)
                else:
                    return x

        data = data.groupby(data.columns[0]).transform(_interpolWrapper)
        # squeezing the 1-dimensional frame resulting from groupby for consistency reasons
        data = data.squeeze(axis=1)
        data.name = dat_name
    return data, chunk_bounds


def _reshapeFlags(
    flagger,
    field,
    ref_index,
    method="fshift",
    agg_method=max,
    missing_flag=None,
    set_shift_comment=True,
    block_flags=None,
    **kwargs,
):
    """To continue processing flags after harmonization/interpolation, old pre-harm flags have to be distributed onto
    new grid.

    There are the following methods available for flags projection. Note, that not every combination of flags projection
    and interpolation method will lead to useful results. (For example, interpolating with 'fshift' and projecting with
    bfill' would be a bad approach obviously.):

    Passed aggregation methods shall return a valid flag for empty sampling intervals, or the value np.nan
    - since np.nan values will be replaced by "missing_flag" anyway.

    'fshift'/'bshift'   - forward/backward projection. Only the very
                        first/last flag will be projected onto the last/next grid point. Extra flag fields like comment,
                        just get shifted along with the flag. Only inserted flags for empty intervals will take the
                        **kwargs argument.
                        Set 'set_shift_comment' to True,  to apply kwargs** to all flags (currently really slow)
    'fagg'/'bagg'       - All flags, referring to a sampling intervals measurements get aggregated forward/backward
                        with the agg_method selected.

    'nshift'     - every grid point gets assigned the nearest flag in its range
                        ( range = grid_point +/-(freq/2) ).Extra flag fields like comment,
                        just get shifted along with the flag. Only inserted flags for empty intervals will take the
                        **kwargs argument.

    'nagg'         - every grid point gets assigned the aggregation (by agg_method), of all the flags in its range.

    :param flagger:     saqc.flagger. The flagger, the passed flags frame refer to.
    :param method:      String. Default = 'fshift'. A methods keyword. (see func doc above)
    :param agg_method:  Func. Default = max. method, multiple flags shall be aggregated with, if an aggregation method
                        is selected for flags projection.
    :param missing_flag:Integer. Default = -1. If there were no flags referring to the harmonized interval, this
                        parameter determines wich flag will be inserted into the reshaped flags frame by selecting
                        flagger.flags[missing_flag]. The parameter defaults to the worst flag that can be thought of, in
                        terms of the used flagger.
    :param set_shift_comment:   Boolean. Default = False. The shifting methods for flags projection are really fast,
                        however, the methods used, do not allow to 'reflag' and apply eventually passed **kwargs.
                        Setting set_shift_comment to True, **kwargs will be applied, but the whole process will slow
                        down significantly.
    :block_flags:       DatetimeIndex. A DatetimeIndex containing labels that will get the "nan-flag" assigned.
                        This option mainly is introduced to account for the backtracking inconsistencies at the end
                        and beginning of interpolation chunks.
    :return: flags:     pd.Series/pd.DataFrame. The reshaped pandas like Flags object, referring to the harmonized data.
    """

    missing_flag = missing_flag or flagger.BAD
    aggregations = [
        "nagg",
        "bagg",
        "fagg",
        "nagg_no_deharm",
        "bagg_no_deharm",
        "fagg_no_deharm",
    ]
    shifts = ["fshift", "bshift", "nshift"]

    freq = ref_index.freqstr

    if method in shifts:
        # forward/backward projection of every intervals last/first flag - rest will be dropped
        if method == "fshift":
            direction = "ffill"
            tolerance = pd.Timedelta(freq)

        elif method == "bshift":
            direction = "bfill"
            tolerance = pd.Timedelta(freq)
        # varset for nshift
        else:
            direction = "nearest"
            tolerance = pd.Timedelta(freq) / 2

        flags = flagger.getFlags().reindex(ref_index, tolerance=tolerance, method=direction, fill_value=np.nan)

        # if you want to keep previous comments - only newly generated missing flags get commented:
        flags_series = flags.squeeze()

        flagger_new = flagger.initFlags(flags=flags).setFlags(
            field, loc=flags_series.isna(), flag=missing_flag, force=True, **kwargs
        )

        if set_shift_comment:
            flagger_new = flagger_new.setFlags(field, flag=flags_series, force=True, **kwargs)

    elif method in aggregations:
        # prepare resampling keywords
        if method in ["fagg", "fagg_no_deharm"]:
            closed = "right"
            label = "right"
            base = 0
            freq_string = freq
        elif method in ["bagg", "bagg_no_deharm"]:
            closed = "left"
            label = "left"
            base = 0
            freq_string = freq
        # var sets for 'nagg':
        else:
            closed = "left"
            label = "left"
            seconds_total = pd.Timedelta(freq).total_seconds()
            base = seconds_total / 2
            freq_string = str(int(seconds_total)) + "s"
            i_start = flagger.getFlags().index[0]
            if abs(i_start - i_start.floor(freq)) <= pd.Timedelta(freq) / 2:
                shift_correcture = 1
            else:
                shift_correcture = -1

        # resampling the flags series with aggregation method
        flags = (
            flagger.getFlags()
            # NOTE: otherwise the datetime index will get lost
            .squeeze()
            .resample(freq_string, closed=closed, label=label, base=base)
            # NOTE: breaks for non categorical flaggers
            .apply(lambda x: agg_method(x) if not x.empty else missing_flag)
            .astype(flagger.dtype)
        )

        if method == "nagg":
            flags = flags.shift(periods=shift_correcture, freq=pd.Timedelta(freq) / 2)

        # some consistency clean up to ensure new flags frame matching new data frames size:
        if ref_index[0] != flags.index[0]:
            flags = pd.Series(data=flagger.BAD, index=[ref_index[0]]).astype(flagger.dtype).append(flags)
        if ref_index[-1] != flags.index[-1]:
            flags = flags.append(pd.Series(data=flagger.BAD, index=[ref_index[-1]]).astype(flagger.dtype))

        # block flagging/backtracking of chunk_starts/chunk_ends
        if block_flags is not None:
            flags[block_flags] = np.nan

        flagger_new = flagger.initFlags(flags=flags.to_frame(name=field))

    else:
        methods = ", ".join(shifts + ["\n"] + aggregations)
        raise ValueError(
            "Passed reshaping method keyword:'{}', is unknown. Please select from: \n '{}'.".format(method, methods)
        )

    # block flagging/backtracking of chunk_starts/chunk_ends
    if block_flags is not None:
        flagger_new = flagger_new.setFlags(
            field, loc=block_flags, flag=pd.Series(np.nan, index=block_flags).astype(flagger_new.dtype), force=True,
        )
    return flagger_new


def _backtrackFlags(flagger_post, flagger_pre, freq, track_method="invert_fshift", co_flagging=False):

    # in the case of "real" up/downsampling - evaluating the harm flags against the original flags makes no sence!
    if track_method in ["regain"]:
        return flagger_pre

    # NOTE: PROBLEM flager_pre carries one value ib exces (index: -3)
    flags_post = flagger_post.getFlags()
    flags_pre = flagger_pre.getFlags()
    flags_header = flags_post.columns
    if track_method in ["invert_fshift", "invert_bshift", "invert_nearest"] and co_flagging is True:
        if track_method == "invert_fshift":
            method = "bfill"
            tolerance = pd.Timedelta(freq)
        elif track_method == "invert_bshift":
            method = "ffill"
            tolerance = pd.Timedelta(freq)
        # var set for "invert nearest"
        else:
            # NOTE: co_flagging bug path
            method = "nearest"
            tolerance = pd.Timedelta(freq) / 2

        flags_post = flags_post.reindex(flags_pre.index, method=method, tolerance=tolerance)
        replacement_mask = flags_post.squeeze() > flags_pre.squeeze()
        # there is a mysterious problem when assigning 1-d-dataframes - so we squeeze:
        flags_pre = flags_pre.squeeze(axis=1)
        flags_post = flags_post.squeeze(axis=1)
        flags_pre.loc[replacement_mask] = flags_post.loc[replacement_mask]

    if track_method in ["invert_fshift", "invert_bshift", "invert_nearest"] and co_flagging is False:
        if track_method == "invert_fshift":
            method = "backward"
            tolerance = pd.Timedelta(freq)
        elif track_method == "invert_bshift":
            method = "forward"
            tolerance = pd.Timedelta(freq)
        # var set for 'invert nearest'
        else:
            method = "nearest"
            tolerance = pd.Timedelta(freq) / 2

        flags_post = pd.merge_asof(
            flags_post,
            pd.DataFrame(flags_pre.index.values, index=flags_pre.index, columns=["pre_index"]),
            left_index=True,
            right_index=True,
            tolerance=tolerance,
            direction=method,
        )

        flags_post.dropna(subset=["pre_index"], inplace=True)
        flags_post.set_index(["pre_index"], inplace=True)

        # restore flag shape
        flags_post.columns = flags_header

        replacement_mask = flags_post.squeeze() > flags_pre.loc[flags_post.index, :].squeeze()
        # there is a mysterious problem when assigning 1-d-dataframes - so we squeeze:
        flags_pre = flags_pre.squeeze(axis=1)
        flags_post = flags_post.squeeze(axis=1)
        flags_pre.loc[replacement_mask[replacement_mask].index] = flags_post.loc[replacement_mask]

    # sticking to the nomenklatura of always-DF for flags:
    if isinstance(flags_pre, pd.Series):
        flags_pre = flags_pre.to_frame()

    return flagger_pre.initFlags(flags=flags_pre)


def _fromMerged(data, flagger, fieldname):
    # we need a not-na mask for the flags data to be retrieved:
    mask = flagger.getFlags(fieldname).notna()
    return data.loc[mask, fieldname], flagger.getFlagger(field=fieldname, loc=mask)


def _toMerged(data, flagger, fieldname, data_to_insert, flagger_to_insert, target_index=None):

    data = data.copy()
    flags = flagger.getFlags()
    flags_to_insert = flagger_to_insert.getFlags()

    if isinstance(data, pd.Series):
        data = data.to_frame()

    data.drop(fieldname, axis="columns", errors="ignore", inplace=True)
    flags.drop(fieldname, axis="columns", errors="ignore", inplace=True)

    # first case: there is no data, the data-to-insert would have
    # to be merged with, and also are we not deharmonizing:
    if (data.empty) and (target_index is None):
        return data_to_insert.to_frame(name=fieldname), flagger_to_insert

    # if thats not the case: generate the drop mask for the remaining data:
    mask = data.isna().all(axis=1)
    # we only want to drop lines, that do not have to be re-inserted in the merge:
    drops = mask[mask].index.difference(data_to_insert.index)
    # clear mask, but keep index
    mask[:] = True
    # final mask:
    mask[drops] = False

    # if we are not "de-harmonizing":
    if target_index is None:
        # erase nan rows in the data, that became redundant because of harmonization and merge with data-to-insert:
        data = pd.merge(data[mask], data_to_insert, how="outer", left_index=True, right_index=True)
        flags = pd.merge(flags[mask], flags_to_insert, how="outer", left_index=True, right_index=True)
        return data, flagger.initFlags(flags=flags)

    else:
        # trivial case: there is only one variable ("reindexing to make sure shape matches pre-harm shape"):
        if data.empty:
            data = data_to_insert.reindex(target_index).to_frame(name=fieldname)
            flags = flags_to_insert.reindex(target_index, fill_value=flagger.UNFLAGGED)
            return data, flagger.initFlags(flags=flags)
        # annoying case: more than one variables:
        # erase nan rows resulting from harmonization but keep/regain those, that were initially present in the data:
        new_index = data[mask].index.join(target_index, how="outer")
        data = data.reindex(new_index)
        flags = flags.reindex(new_index, fill_value=flagger.UNFLAGGED)
        data = pd.merge(data, data_to_insert, how="outer", left_index=True, right_index=True)
        flags = pd.merge(flags, flags_to_insert, how="outer", left_index=True, right_index=True)

        # internally harmonization memorizes its own manipulation by inserting nan flags -
        # those we will now assign the flagger.bad flag by the "missingTest":
        return flagMissing(data, fieldname, flagger.initFlags(flags=flags), nodata=np.nan)


@register("harmonize_shift2Grid")
def shift2Grid(data, field, flagger, freq, method="nshift", drop_flags=None, **kwargs):
    return harmonize(
        data, field, flagger, freq, inter_method=method, reshape_method=method, drop_flags=drop_flags, **kwargs,
    )


@register("harmonize_aggregate2Grid")
def aggregate2Grid(
    data, field, flagger, freq, value_func, flag_func="max", method="nagg", drop_flags=None, **kwargs,
):
    return harmonize(
        data,
        field,
        flagger,
        freq,
        inter_method=method,
        reshape_method=method,
        inter_agg=value_func,
        reshape_agg=flag_func,
        drop_flags=drop_flags,
        **kwargs,
    )


@register("harmonize_linear2Grid")
def linear2Grid(data, field, flagger, freq, method="nagg", func="max", drop_flags=None, **kwargs):
    return harmonize(
        data,
        field,
        flagger,
        freq,
        inter_method="time",
        reshape_method=method,
        reshape_agg=func,
        drop_flags=drop_flags,
        **kwargs,
    )


@register("harmonize_interpolate2Grid")
def interpolate2Grid(
    data, field, flagger, freq, method, order=1, flag_method="nagg", flag_func="max", drop_flags=None, **kwargs,
):
    return harmonize(
        data,
        field,
        flagger,
        freq,
        inter_method=method,
        inter_order=order,
        reshape_method=flag_method,
        reshape_agg=flag_func,
        drop_flags=drop_flags,
        **kwargs,
    )


@register("harmonize_downsample")
def downsample(
    data,
    field,
    flagger,
    sample_freq,
    agg_freq,
    sample_func="mean",
    agg_func="mean",
    invalid_flags=None,
    max_invalid=None,
    **kwargs,
):

    agg_func = getFuncFromInput(agg_func)

    if max_invalid is None:
        max_invalid = np.inf

    if sample_func is not None:
        sample_func = getFuncFromInput(sample_func)

    # define the "fastest possible" aggregator
    if sample_func is None:
        if max_invalid < np.inf:

            def aggregator(x):
                if x.isna().sum() < max_invalid:
                    return agg_func(x)
                else:
                    return np.nan

        else:

            def aggregator(x):
                return agg_func(x)

    else:

        dummy_resampler = pd.Series(np.nan, index=[pd.Timedelta("1min")]).resample("1min")
        if hasattr(dummy_resampler, sample_func.__name__):

            sample_func_name = sample_func.__name__
            if max_invalid < np.inf:

                def aggregator(x):
                    y = getattr(x.resample(sample_freq), sample_func_name)()
                    if y.isna().sum() < max_invalid:
                        return agg_func(y)
                    else:
                        return np.nan

            else:

                def aggregator(x):
                    return agg_func(getattr(x.resample(sample_freq), sample_func_name)())

        else:
            if max_invalid < np.inf:

                def aggregator(x):
                    y = x.resample(sample_freq).apply(sample_func)
                    if y.isna().sum() < max_invalid:
                        return agg_func(y)
                    else:
                        return np.nan

            else:

                def aggregator(x):
                    return agg_func(x.resample(sample_freq).apply(sample_func))

    return harmonize(
        data,
        field,
        flagger,
        agg_freq,
        inter_method="bagg",
        reshape_method="bagg_no_deharm",
        inter_agg=aggregator,
        reshape_agg="max",
        drop_flags=invalid_flags,
        **kwargs,
    )



import pytest

import pandas as pd

from saqc.funcs.machine_learning import flagML

from test.common import TESTFLAGGER


@pytest.mark.parametrize("flagger", TESTFLAGGER)
def test_flagML(flagger):
    ### CREATE MWE DATA
    data = pd.read_feather("ressources/machine_learning/data/soil_moisture_mwe.feather")
    data = data.set_index(pd.DatetimeIndex(data.Time))
    flags_raw = data[["SM1_Flag", "SM2_Flag", "SM3_Flag"]]
    flags_raw.columns = ["SM1", "SM2", "SM3"]

    # masks for flag preparation
    mask_bad = flags_raw.isin(["Auto:BattV", "Auto:Range", "Auto:Spike"])
    mask_unflagged = flags_raw.isin(["Manual"])
    mask_good = flags_raw.isin(["OK"])

    field = "SM2"

    # prepare flagsframe
    flagger = flagger.initFlags(data)
    flagger = flagger.setFlags(field, loc=mask_bad[field])
    flagger = flagger.setFlags(field, loc=mask_unflagged[field], flag=flagger.UNFLAGGED)
    flagger = flagger.setFlags(field, loc=mask_good[field], flag=flagger.GOOD)

    references = ["Temp2", "BattV"]
    window_values = 20
    window_flags = 20
    groupvar = 0.2
    modelname = "testmodel"
    path = f"ressources/machine_learning/models/{modelname}_{groupvar}.pkl"

    outdat, outflagger = flagML(
        data, field, flagger, references, window_values, window_flags, path
    )

    # compare
    # assert resulting no of bad flags
    badflags = outflagger.isFlagged(field)
    assert badflags.sum() == 10447

    # Have the right values been flagged?
    checkdates = pd.DatetimeIndex(
        [
            "2014-08-05 23:03:59",
            "2014-08-06 01:35:44",
            "2014-08-06 01:50:54",
            "2014-08-06 02:06:05",
            "2014-08-06 02:21:15",
            "2014-08-06 04:22:38",
            "2014-08-06 04:37:49",
            "2014-08-06 04:52:59",
        ]
    )
    assert badflags[checkdates].all()

    # IN CASE OF FUTURE changes
    # Get indices of values that were flagged
    # wasmanual = flags_raw[field]=="Manual"
    # a = badflags & wasmanual
    # data.loc[a,:].index



#!/usr/bin/env python

__author__ = "Bert Palm"
__email__ = "bert.palm@ufz.de"
__copyright__ = "Copyright 2018, Helmholtz-Zentrum für Umweltforschung GmbH - UFZ"

import pytest
import numpy as np
import pandas as pd
from pandas.api.types import is_bool_dtype

from test.common import TESTFLAGGER


def _getDataset(rows, cols):
    df = pd.DataFrame()
    for c in range(cols):
        df[f"var{c}"] = np.linspace(0 + 100 * c, rows, rows)
    vals = pd.date_range(start="2011-01-01", end="2011-01-10", periods=rows)
    df.index = pd.DatetimeIndex(data=vals)
    return df


DATASETS = [
    # _getDataset(0, 1),
    # _getDataset(1, 1),
    _getDataset(100, 1),
    # _getDataset(1000, 1),
    # _getDataset(0, 4),
    # _getDataset(1, 4),
    _getDataset(100, 4),
    # _getDataset(1000, 4),
    # _getDataset(10000, 40),
    # _getDataset(20, 4),
]


@pytest.mark.parametrize("data", DATASETS)
@pytest.mark.parametrize("flagger", TESTFLAGGER)
def test_setFlagger(data, flagger):

    field, *_ = data.columns

    this_flagger = flagger.initFlags(data)
    other_flagger = this_flagger.getFlagger(iloc=slice(None, None, 3)).setFlags(field)
    result_flagger = this_flagger.setFlagger(other_flagger)

    other_flags = other_flagger.getFlags()
    result_flags = result_flagger.getFlags(field)

    assert np.all(
        result_flagger.getFlags(loc=other_flagger.getFlags().index) == other_flags
    )

    assert np.all(
        result_flags[~result_flags.index.isin(other_flags.index)] == flagger.UNFLAGGED
    )


@pytest.mark.parametrize("data", DATASETS)
@pytest.mark.parametrize("flagger", TESTFLAGGER)
def test_setFlaggerColumnsDiff(data, flagger):

    field, *_ = data.columns
    new_field = field + "_new"
    iloc = slice(None, None, 2)

    other_data = data.iloc[iloc]
    other_data.columns = [new_field] + data.columns[1:].to_list()

    this_flagger = flagger.initFlags(data).setFlags(field, flag=flagger.BAD)
    other_flagger = flagger.initFlags(other_data)
    result_flagger = this_flagger.setFlagger(other_flagger)

    assert np.all(
        result_flagger.getFlags(new_field, loc=other_data.index)
        == other_flagger.getFlags(new_field)
    )
    assert np.all(
        result_flagger.getFlags(new_field, loc=data.index) == flagger.UNFLAGGED
    )


@pytest.mark.parametrize("data", DATASETS)
@pytest.mark.parametrize("flagger", TESTFLAGGER)
def test_setFlaggerIndexDiff(data, flagger):

    field, *_ = data.columns
    iloc = slice(None, None, 2)

    other_data = data.iloc[iloc]
    other_data.index = other_data.index + pd.Timedelta(minutes=2, seconds=25)

    this_flagger = flagger.initFlags(data).setFlags(field, flag=flagger.BAD)
    other_flagger = flagger.initFlags(other_data)
    result_flagger = this_flagger.setFlagger(other_flagger)

    assert np.all(
        result_flagger.getFlags(field, loc=other_data.index)
        == other_flagger.getFlags(field)
    )
    assert np.all(
        result_flagger.getFlags(field, loc=data.index) == this_flagger.getFlags(field)
    )


@pytest.mark.parametrize("data", DATASETS)
@pytest.mark.parametrize("flagger", TESTFLAGGER)
def test_initFlags(data, flagger):
    flags = flagger.initFlags(data).getFlags()
    assert isinstance(flags, pd.DataFrame)
    assert len(flags.index) == len(data.index)
    assert len(flags.columns) >= len(data.columns)


@pytest.mark.parametrize("data", DATASETS)
@pytest.mark.parametrize("flagger", TESTFLAGGER)
def test_getFlags(data, flagger):
    flagger = flagger.initFlags(data)
    field, *_ = data.columns

    # df
    flags0 = flagger.getFlags()
    assert isinstance(flags0, pd.DataFrame)
    assert flags0.shape == data.shape
    assert (flags0.columns == data.columns).all()

    for dt in flags0.dtypes:
        assert dt == flagger.dtype

    # series
    flags1 = flagger.getFlags(field)
    assert isinstance(flags1, pd.Series)
    assert flags1.dtype == flagger.dtype
    assert flags1.shape[0] == data.shape[0]
    assert flags1.name in data.columns


@pytest.mark.parametrize("data", DATASETS)
@pytest.mark.parametrize("flagger", TESTFLAGGER)
def test_isFlaggedDataFrame(data, flagger):

    flagger = flagger.initFlags(data)
    field, *_ = data.columns

    mask = np.zeros(len(data), dtype=bool)

    df_tests = [
        (flagger.isFlagged(), mask),
        (flagger.setFlags(field).isFlagged(), ~mask),
        (
            flagger.setFlags(field, flag=flagger.GOOD).isFlagged(
                flag=flagger.GOOD, comparator=">"
            ),
            mask,
        ),
        (
            flagger.setFlags(field, flag=flagger.GOOD).isFlagged(
                flag=flagger.GOOD, comparator="<"
            ),
            mask,
        ),
        (
            flagger.setFlags(field, flag=flagger.GOOD).isFlagged(
                flag=flagger.GOOD, comparator="=="
            ),
            ~mask,
        ),
    ]
    for flags, expected in df_tests:
        assert np.all(flags[field] == expected)
        assert isinstance(flags, pd.DataFrame)
        assert flags.shape == data.shape
        assert (flags.columns == data.columns).all()
        for dt in flags.dtypes:
            assert is_bool_dtype(dt)


@pytest.mark.parametrize("data", DATASETS)
@pytest.mark.parametrize("flagger", TESTFLAGGER)
def test_isFlaggedSeries(data, flagger):

    flagger = flagger.initFlags(data)
    field, *_ = data.columns

    mask = np.zeros(len(data), dtype=bool)

    series_tests = [
        (flagger.isFlagged(field), mask),
        (flagger.setFlags(field).isFlagged(field), ~mask),
        (
            flagger.setFlags(field, flag=flagger.GOOD).isFlagged(
                field, flag=flagger.GOOD, comparator=">"
            ),
            mask,
        ),
        (
            flagger.setFlags(field, flag=flagger.GOOD).isFlagged(
                field, flag=flagger.GOOD, comparator="<"
            ),
            mask,
        ),
        (
            flagger.setFlags(field, flag=flagger.GOOD).isFlagged(
                field, flag=flagger.GOOD, comparator="=="
            ),
            ~mask,
        ),
    ]
    for flags, expected in series_tests:
        assert np.all(flags == expected)
        assert isinstance(flags, pd.Series)
        assert flags.dtype == bool
        assert flags.shape[0] == data.shape[0]
        assert flags.name in data.columns


@pytest.mark.parametrize("data", DATASETS)
@pytest.mark.parametrize("flagger", TESTFLAGGER)
def test_isFlaggedSeries(data, flagger):

    flagger = flagger.initFlags(data)
    field, *_ = data.columns

    fail_tests = [
        {"flag": pd.Series(index=data.index, data=flagger.BAD).astype(flagger.dtype)},
        {"field": ["var1", "var2"]},
    ]
    for args in fail_tests:
        with pytest.raises(ValueError):
            flagger.isFlagged(**args)


@pytest.mark.parametrize("data", DATASETS)
@pytest.mark.parametrize("flagger", TESTFLAGGER)
def test_setFlags(data, flagger):
    flagger = flagger.initFlags(data)
    sl = slice("2011-01-02", "2011-01-05")
    field, *_ = data.columns

    base = flagger.getFlags()

    flagger_good = flagger.setFlags(field, flag=flagger.GOOD, loc=sl)
    flags_good = flagger_good.getFlags()
    assert flags_good.shape == base.shape
    assert (flags_good.columns == base.columns).all()
    assert (flags_good.loc[sl, field] == flagger.GOOD).all()

    # overflag works BAD > GOOD
    flagger_bad = flagger_good.setFlags(field, flag=flagger.BAD)
    assert (flagger_bad.getFlags(field) == flagger.BAD).all()

    # overflag doesn't work GOOD < BAD
    flagger_still_bad = flagger_bad.setFlags(field, flag=flagger.GOOD)
    assert (flagger_still_bad.getFlags(field) == flagger.BAD).all()

    # overflag does work with force
    flagger_forced_good = flagger_bad.setFlags(field, flag=flagger.GOOD, force=True)
    assert (flagger_forced_good.getFlags(field) == flagger.GOOD).all()

    with pytest.raises(ValueError):
        flagger.setFlags(field=None, flag=flagger.BAD)


@pytest.mark.parametrize("data", DATASETS)
@pytest.mark.parametrize("flagger", TESTFLAGGER)
def test_clearFlags(data, flagger):
    flagger = flagger.initFlags(data)
    sl = slice("2011-01-02", "2011-01-05")
    field, *_ = data.columns

    base = flagger.getFlags()

    flagger = flagger.setFlags(field=field, flag=flagger.BAD)
    assert np.sum(flagger.isFlagged(field)) == len(base)

    flagger = flagger.clearFlags(field)
    assert np.sum(flagger.isFlagged(field)) == 0

    flagger = flagger.setFlags(field=field, flag=flagger.BAD)
    assert np.sum(flagger.isFlagged(field)) == len(base)

    flagger = flagger.clearFlags(field, loc=sl)
    unflagged = flagger.isFlagged(field, loc=sl)
    assert np.sum(unflagged) == 0
    assert np.sum(flagger.isFlagged(field)) == len(data) - len(unflagged)


@pytest.mark.parametrize("data", DATASETS)
@pytest.mark.parametrize("flagger", TESTFLAGGER)
def test_dtype(data, flagger):

    flagger = flagger.initFlags(data)
    field, *_ = data.columns

    tests = (
        flagger.getFlags(field).astype(str),
        "TEST",
        55,
    )

    for test in tests:
        with pytest.raises(TypeError):
            flagger = flagger.setFlags(field, flag=test)
        assert flagger.getFlags(field).dtype == flagger.dtype


@pytest.mark.parametrize("data", DATASETS)
@pytest.mark.parametrize("flagger", TESTFLAGGER[-1:])
def test_returnCopy(data, flagger):

    flagger = flagger.initFlags(data)
    field, *_ = data.columns

    base = flagger.getFlags()

    assert flagger.getFlags() is not base
    assert flagger.isFlagged() is not base
    assert flagger.setFlags(field) is not flagger
    assert flagger.clearFlags(field) is not flagger


LOC_ILOC_FUNCS = ["isFlagged", "getFlags"]


@pytest.mark.parametrize("data", DATASETS)
@pytest.mark.parametrize("flagger", TESTFLAGGER)
@pytest.mark.parametrize("flaggerfunc", LOC_ILOC_FUNCS)
def test_loc(data, flagger, flaggerfunc):

    flagger = flagger.initFlags(data)
    sl = slice("2011-01-02", "2011-01-05")
    field, *_ = data.columns

    chunk = data.loc[sl, field]
    d = data.loc[sl]
    if d.empty:
        mask = []
    else:
        m = data.index.get_loc(d.index[0])
        M = data.index.get_loc(d.index[-1])
        mask = np.full(len(data), False)
        mask[m:M] = True

    flagger_func = getattr(flagger, flaggerfunc)

    # masked
    mflags0 = flagger_func(field, loc=mask)
    mflags1 = flagger_func().loc[mask, field]
    mflags2 = flagger_func(field).loc[mask]
    mflags3 = flagger_func(loc=mask)[field]
    assert (mflags0 == mflags1).all()
    assert (mflags0 == mflags2).all()
    assert (mflags0 == mflags3).all()

    # indexed
    iflags0 = flagger_func(field, loc=chunk.index)
    iflags1 = flagger_func().loc[chunk.index, field]
    iflags2 = flagger_func(field).loc[chunk.index]
    iflags3 = flagger_func(loc=chunk.index)[field]
    assert (iflags0 == iflags1).all()
    assert (iflags0 == iflags2).all()
    assert (iflags0 == iflags3).all()

    # sliced
    sflags0 = flagger_func(field, loc=sl)
    sflags1 = flagger_func().loc[sl, field]
    sflags2 = flagger_func(field).loc[sl]
    sflags3 = flagger_func(loc=sl)[field]
    assert (sflags0 == sflags1).all()
    assert (sflags0 == sflags2).all()
    assert (sflags0 == sflags3).all()

    assert (sflags0 == iflags0).all()


@pytest.mark.parametrize("data", DATASETS)
@pytest.mark.parametrize("flagger", TESTFLAGGER)
@pytest.mark.parametrize("flaggerfunc", LOC_ILOC_FUNCS)
def test_iloc(data, flagger, flaggerfunc):
    flagger = flagger.initFlags(data)
    field, *_ = data.columns

    M = len(data.index) - 1 if len(data.index) > 0 else 0
    m = M // 3
    M = m * 2

    array = data.reset_index(drop=True).index.values[m:M]
    sl = slice(m, M)
    mask = np.full(len(data), False)
    mask[sl] = True

    flagger_func = getattr(flagger, flaggerfunc)

    # masked
    mflags0 = flagger_func(field, iloc=mask)
    mflags1 = flagger_func().iloc[mask, 0]
    mflags2 = flagger_func(field).iloc[mask]
    mflags3 = flagger_func(iloc=mask)[field]
    assert (mflags0 == mflags1).all()
    assert (mflags0 == mflags2).all()
    assert (mflags0 == mflags3).all()

    # indexed
    iflags0 = flagger_func(field, iloc=array)
    iflags1 = flagger_func().iloc[array, 0]
    iflags2 = flagger_func(field).iloc[array]
    iflags3 = flagger_func(iloc=array)[field]
    assert (iflags0 == iflags1).all()
    assert (iflags0 == iflags2).all()
    assert (iflags0 == iflags3).all()

    # sliced
    sflags0 = flagger_func(field, iloc=sl)
    sflags1 = flagger_func().iloc[sl, 0]
    sflags2 = flagger_func(field).iloc[sl]
    sflags3 = flagger_func(iloc=sl)[field]
    assert (sflags0 == sflags1).all()
    assert (sflags0 == sflags2).all()
    assert (sflags0 == sflags3).all()

    assert (sflags0 == iflags0).all()
    assert (sflags0 == mflags0).all()


@pytest.mark.parametrize("data", DATASETS)
@pytest.mark.parametrize("flagger", TESTFLAGGER)
def test_classicUseCases(data, flagger):
    flagger = flagger.initFlags(data)
    field, *_ = data.columns

    # data-mask, same length than flags
    d = data[field]
    mask = d < (d.max() - d.min()) // 2
    flagger = flagger.clearFlags(field)
    flagged = flagger.setFlags(field, loc=mask, flag=flagger.BAD).isFlagged(field)
    assert (flagged == mask).all()

    # some fun with numpy but not same dimensions.. pass indices to iloc
    indices = np.arange(0, len(data))
    mask = indices % 3 == 0
    indices = indices[mask]
    flagger.clearFlags(field)
    flagged = flagger.setFlags(field, iloc=indices, flag=flagger.BAD).isFlagged(field)
    assert (flagged.iloc[indices] == flagged[flagged]).all()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import logging

import numpy as np
import pandas as pd
from scipy.signal import savgol_filter

from saqc.funcs.break_detection import flagBreaksSpektrumBased
from saqc.funcs.spike_detection import flagSpikes_spektrumBased
from saqc.funcs.constants_detection import flagConstantVarianceBased
from saqc.funcs.register import register
from saqc.lib.tools import retrieveTrustworthyOriginal


@register("soilMoisture_spikes")
def flagSoilmoistureSpikes(
    data,
    field,
    flagger,
    raise_factor=0.15,
    deriv_factor=0.2,
    noise_func="CoVar",
    noise_window="12h",
    noise_thresh=1,
    smooth_window="3h",
    smooth_poly_deg=2,
    **kwargs
):

    """
    The Function provides just a call to flagSpikes_spektrumBased, with parameter defaults, that refer to:

    Dorigo,W,.... Global Automated Quality Control of In Situ Soil Moisture Data from the international
    Soil Moisture Network. 2013. Vadoze Zone J. doi:10.2136/vzj2012.0097.
    """

    return flagSpikes_spektrumBased(
        data,
        field,
        flagger,
        raise_factor=raise_factor,
        deriv_factor=deriv_factor,
        noise_func=noise_func,
        noise_window=noise_window,
        noise_thresh=noise_thresh,
        smooth_window=smooth_window,
        smooth_poly_deg=smooth_poly_deg,
        **kwargs
    )


@register("soilMoisture_breaks")
def flagSoilMoistureBreaks(
    data,
    field,
    flagger,
    thresh_rel=0.1,
    thresh_abs=0.01,
    first_der_factor=10,
    first_der_window="12h",
    scnd_der_ratio_range=0.05,
    scnd_der_ratio_thresh=10,
    smooth=False,
    smooth_window="3h",
    smooth_poly_deg=2,
    **kwargs
):

    """
    The Function provides just a call to flagBreaks_spektrumBased, with parameter defaults that refer to:

    Dorigo,W,.... Global Automated Quality Control of In Situ Soil Moisture Data from the international
    Soil Moisture Network. 2013. Vadoze Zone J. doi:10.2136/vzj2012.0097.

    """
    return flagBreaksSpektrumBased(
        data,
        field,
        flagger,
        thresh_rel=thresh_rel,
        thresh_abs=thresh_abs,
        first_der_factor=first_der_factor,
        first_der_window=first_der_window,
        scnd_der_ratio_range=scnd_der_ratio_range,
        scnd_der_ratio_thresh=scnd_der_ratio_thresh,
        smooth=smooth,
        smooth_window=smooth_window,
        smooth_poly_deg=smooth_poly_deg,
        **kwargs
    )


@register("soilMoisture_frost")
def flagSoilMoistureBySoilFrost(data, field, flagger, soil_temp_variable, window="1h", frost_thresh=0, **kwargs):

    """This Function is an implementation of the soil temperature based Soil Moisture flagging, as presented in:

    Dorigo,W,.... Global Automated Quality Control of In Situ Soil Moisture Data from the international
    Soil Moisture Network. 2013. Vadoze Zone J. doi:10.2136/vzj2012.0097.

    All parameters default to the values, suggested in this publication.

    Function flags Soil moisture measurements by evaluating the soil-frost-level in the moment of measurement.
    Soil temperatures below "frost_level" are regarded as denoting frozen soil state.

    :param data:                        The pandas dataframe holding the data-to-be flagged, as well as the reference
                                        series. Data must be indexed by a datetime series.
    :param field:                       Fieldname of the Soil moisture measurements field in data.
    :param flagger:                     A flagger - object.
                                        like thingies that refer to the data(including datestrings).
    :param tolerated_deviation:         Offset String. Denoting the maximal temporal deviation,
                                        the soil frost states timestamp is allowed to have, relative to the
                                        data point to-be-flagged.
    :param soil_temp_reference:         A STRING, denoting the fields name in data,
                                        that holds the data series of soil temperature values,
                                        the to-be-flagged values shall be checked against.
    :param frost_level:                 Value level, the flagger shall check against, when evaluating soil frost level.
    """

    # retrieve reference series
    refseries = data[soil_temp_variable].copy()
    ref_use = flagger.isFlagged(soil_temp_variable, flag=flagger.GOOD, comparator="==") | flagger.isFlagged(
        soil_temp_variable, flag=flagger.UNFLAGGED, comparator="=="
    )
    # drop flagged values:
    refseries = refseries[ref_use.values]
    # drop nan values from reference series, since those are values you dont want to refer to.
    refseries = refseries.dropna()
    # skip further processing if reference series is empty:
    if refseries.empty:
        return data, flagger

    refseries = refseries.reindex(data[field].dropna().index, method="nearest", tolerance=window)
    refseries = refseries[refseries < frost_thresh].index

    flagger = flagger.setFlags(field, refseries, **kwargs)
    return data, flagger


@register("soilMoisture_precipitation")
def flagSoilMoistureByPrecipitationEvents(
    data,
    field,
    flagger,
    prec_variable,
    raise_window=None,
    sensor_depth=0,
    sensor_accuracy=0,
    soil_porosity=0,
    std_factor=2,
    std_window="24h",
    ignore_missing=False,
    **kwargs
):

    """This Function is an implementation of the precipitation based Soil Moisture flagging, as presented in:

    Dorigo,W,.... Global Automated Quality Control of In Situ Soil Moisture Data from the international
    Soil Moisture Network. 2013. Vadoze Zone J. doi:10.2136/vzj2012.0097.

    All parameters default to the values, suggested in this publication. (excluding porosity,sensor accuracy and
    sensor depth)


    Function flags Soil moisture measurements by flagging moisture rises that do not follow up a sufficient
    precipitation event. If measurement depth, sensor accuracy of the soil moisture sensor and the porosity of the
    surrounding soil is passed to the function, an inferior level of precipitation, that has to preceed a significant
    moisture raise within 24 hours, can be estimated. If those values are not delivered, this inferior bound is set
    to zero. In that case, any non zero precipitation count will justify any soil moisture raise.

    A data point y_t is flagged an invalid soil moisture raise, if:

    (1) y_t > y_(t-raise_window)
    (2) y_t - y_(t-"std_factor_range") > "std_factor" * std(y_(t-"std_factor_range"),...,y_t)
    (3) sum(prec(t-24h),...,prec(t)) > sensor_depth * sensor_accuracy * soil_porosity

    NOTE1: np.nan entries in the input precipitation series will be regarded as susipicious and the test will be
    omited for every 24h interval including a np.nan entrie in the original precipitation sampling rate.
    Only entry "0" will be regarded as denoting "No Rainfall".

    NOTE2: The function wont test any values that are flagged suspicious anyway - this may change in a future version.


    :param data:                        The pandas dataframe holding the data-to-be flagged, as well as the reference
                                        series. Data must be indexed by a datetime series and be harmonized onto a
                                        time raster with seconds precision.
    :param field:                       Fieldname of the Soil moisture measurements field in data.
    :param flagger:                     A flagger - object. (saqc.flagger.X)
    :param prec_variable:               Fieldname of the precipitation meassurements column in data.
    :param sensor_depth:                Measurement depth of the soil moisture sensor, [m].
    :param sensor_accuracy:             Accuracy of the soil moisture sensor, [-].
    :param soil_porosity:               Porosity of moisture sensors surrounding soil, [-].
    :param std_factor:                  The value determines by which rule it is decided, weather a raise in soil
                                        moisture is significant enough to trigger the flag test or not:
                                        Significants is assumed, if the raise is  greater then "std_factor" multiplied
                                        with the last 24 hours standart deviation.
    :param std_factor_range:            Offset String. Denotes the range over witch the standart deviation is obtained,
                                        to test condition [2]. (Should be a multiple of the sampling rate)
    :param raise_window:                Offset String. Denotes the distance to the datapoint, relatively to witch
                                        it is decided if the current datapoint is a raise or not. Equation [1].
                                        It defaults to None. When None is passed, raise_window is just the sample
                                        rate of the data. Any raise reference must be a multiple of the (intended)
                                        sample rate and below std_factor_range.
    :param ignore_missing:
    """

    dataseries, moist_rate = retrieveTrustworthyOriginal(data, field, flagger)

    # data not hamronized:
    refseries = data[prec_variable].dropna()
    # abort processing if any of the measurement series has no valid entries!
    if moist_rate is np.nan:
        return data, flagger
    if refseries.empty:
        return data, flagger

    refseries = refseries.reindex(refseries.index.join(dataseries.index, how="outer"))
    # get 24 h prec. monitor
    prec_count = refseries.rolling(window="1D").sum()
    # exclude data not signifying a raise::
    if raise_window is None:
        raise_window = 1
    else:
        raise_window = int(np.ceil(pd.Timedelta(raise_window) / moist_rate))

    # first raise condition:
    raise_mask = dataseries > dataseries.shift(raise_window)

    # second raise condition:
    std_window = int(np.ceil(pd.Timedelta(std_window) / moist_rate))
    if ignore_missing:
        std_mask = dataseries.dropna().rolling(std_window).std() < (
            (dataseries - dataseries.shift(std_window)) / std_factor
        )
    else:
        std_mask = dataseries.rolling(std_window).std() < ((dataseries - dataseries.shift(std_window)) / std_factor)

    dataseries = dataseries[raise_mask & std_mask]
    invalid_indices = prec_count[dataseries.index] <= sensor_depth * sensor_accuracy * soil_porosity
    invalid_indices = invalid_indices[invalid_indices]

    # set Flags
    flagger = flagger.setFlags(field, loc=invalid_indices.index, **kwargs)
    return data, flagger


@register("soilMoisture_constant")
def flagSoilMoistureConstant(
    data,
    field,
    flagger,
    window="12h",
    thresh=0.0005,
    precipitation_window="12h",
    tolerance=0.95,
    deriv_max=0.0025,
    deriv_min=0,
    max_missing=None,
    max_consec_missing=None,
    smooth_window=None,
    smooth_poly_deg=2,
    **kwargs
):

    """

    Note, function has to be harmonized to equidistant freq_grid

    Note, in current implementation, it has to hold that: (rainfall_window_range >= plateau_window_min)

    :param data:                        The pandas dataframe holding the data-to-be flagged.
                                        Data must be indexed by a datetime series and be harmonized onto a
                                        time raster with seconds precision (skips allowed).
    :param field:                       Fieldname of the Soil moisture measurements field in data.
    :param flagger:                     A flagger - object. (saqc.flagger.X)
    """

    # get plateaus:
    _, comp_flagger = flagConstantVarianceBased(
        data,
        field,
        flagger,
        window=window,
        thresh=thresh,
        max_missing=max_missing,
        max_consec_missing=max_consec_missing,
    )

    new_plateaus = (comp_flagger.getFlags(field)).eq(flagger.getFlags(field))
    # get dataseries at its sampling freq:
    dataseries, moist_rate = retrieveTrustworthyOriginal(data, field, flagger)
    # get valuse referring to dataseries:
    new_plateaus.resample(pd.Timedelta(moist_rate)).asfreq()
    # cut out test_slices for min/max derivatives condition check:
    # offset 2 periods:
    precipitation_window = int(np.ceil(pd.Timedelta(precipitation_window) / moist_rate))
    window = int(np.ceil(pd.Timedelta(window) / moist_rate))
    period_diff = precipitation_window - window
    # we cast plateua series to int - because replace has problems with replacing bools by "method".
    new_plateaus = new_plateaus.astype(int)
    # get plateau groups:
    group_counter = new_plateaus.cumsum()
    group_counter = group_counter[group_counter.diff() == 0]
    group_counter.name = "group_counter"
    plateau_groups = pd.merge(group_counter, dataseries, left_index=True, right_index=True, how="inner")
    # test mean-condition on plateau groups:
    test_barrier = tolerance * dataseries.max()
    plateau_group_drops = plateau_groups.groupby("group_counter").filter(lambda x: x[field].mean() <= test_barrier)
    # discard values that didnt pass the test from plateau candidate series:
    new_plateaus[plateau_group_drops.index] = 1

    # we extend the plateaus to cover condition testing sets
    # 1: extend backwards (with a technical "one" added):
    cond1_sets = new_plateaus.replace(1, method="bfill", limit=(precipitation_window + window))
    # 2. extend forwards:
    if period_diff > 0:
        cond1_sets = cond1_sets.replace(1, method="ffill", limit=period_diff)

    # get first derivative
    if smooth_window is None:
        smooth_window = 3 * pd.Timedelta(moist_rate)
    else:
        smooth_window = pd.Timedelta(smooth_window)
    first_derivative = dataseries.diff()
    filter_window_seconds = smooth_window.seconds
    smoothing_periods = int(np.ceil((filter_window_seconds / moist_rate.n)))
    first_derivate = savgol_filter(dataseries, window_length=smoothing_periods, polyorder=smooth_poly_deg, deriv=1,)
    first_derivate = pd.Series(data=first_derivate, index=dataseries.index, name=dataseries.name)
    # cumsumming to seperate continous plateau groups from each other:
    group_counter = cond1_sets.cumsum()
    group_counter = group_counter[group_counter.diff() == 0]
    group_counter.name = "group_counter"
    group_frame = pd.merge(group_counter, first_derivate, left_index=True, right_index=True, how="inner")
    group_frame = group_frame.groupby("group_counter")
    condition_passed = group_frame.filter(lambda x: (x[field].max() >= deriv_max) & (x[field].min() <= deriv_min))

    flagger = flagger.setFlags(field, loc=condition_passed.index, **kwargs)

    return data, flagger



import pandas as pd
import numpy as np
from saqc.funcs.register import register

# NEW
from sklearn.ensemble import RandomForestClassifier
import joblib


def _refCalc(reference, window_values):
    # Helper function for calculation of moving window values
    outdata = pd.DataFrame()
    name = reference.name
    # derive gradients from reference series
    outdata[name + "_Dt_1"] = reference - reference.shift(1)  # gradient t vs. t-1
    outdata[name + "_Dt1"] = reference - reference.shift(-1)  # gradient t vs. t+1
    # moving mean of gradients var1 and var2 before/after
    outdata[name + "_Dt_" + str(window_values)] = (
        outdata[name + "_Dt_1"].rolling(window_values, center=False).mean()
    )  # mean gradient t to t-window
    outdata[name + "_Dt" + str(window_values)] = (
        outdata[name + "_Dt_1"].iloc[::-1].rolling(window_values, center=False).mean()[::-1]
    )  # mean gradient t to t+window
    return outdata


@register("machinelearning")
def flagML(data, field, flagger, references, window_values: int, window_flags: int, path: str, **kwargs):

    """This Function uses pre-trained machine-learning model objects for flagging of a specific variable. The model is supposed to be trained using the script provided in "ressources/machine_learning/train_machine_learning.py".
    For flagging, Inputs to the model are the timeseries of the respective target at one specific sensors, the automatic flags that were assigned by SaQC as well as multiple reference series.
    Internally, context information for each point is gathered in form of moving windows to improve the flagging algorithm according to user input during model training.
    For the model to work, the parameters 'references', 'window_values' and 'window_flags' have to be set to the same values as during training.
    :param data:                        The pandas dataframe holding the data-to-be flagged, as well as the reference series. Data must be indexed by a datetime index.
    :param flags:                       A dataframe holding the flags
    :param field:                       Fieldname of the field in data that is to be flagged.
    :param flagger:                     A flagger - object.
    :param references:                  A string or list of strings, denoting the fieldnames of the data series that should be used as reference variables
    :param window_values:               An integer, denoting the window size that is used to derive the gradients of both the field- and reference-series inside the moving window
    :param window_flags:                An integer, denoting the window size that is used to count the surrounding automatic flags that have been set before
    :param path:                        A string giving the path to the respective model object, i.e. its name and the respective value of the grouping variable. e.g. "models/model_0.2.pkl"
    """

    # Function for moving window calculations
    # Create custom df for easier processing
    df = data.loc[:, [field] + references]
    # Create binary column of BAD-Flags
    df["flag_bin"] = flagger.isFlagged(field, flag=flagger.BAD, comparator="==").astype(
        "int"
    )  # get "BAD"-flags and turn into binary

    # Add context information of flags
    df["flag_bin_t_1"] = df["flag_bin"] - df["flag_bin"].shift(1)  # Flag at t-1
    df["flag_bin_t1"] = df["flag_bin"] - df["flag_bin"].shift(-1)  # Flag at t+1
    df["flag_bin_t_" + str(window_flags)] = (
        df["flag_bin"].rolling(window_flags + 1, center=False).sum()
    )  # n Flags in interval t to t-window_flags
    df["flag_bin_t" + str(window_flags)] = (
        df["flag_bin"].iloc[::-1].rolling(window_flags + 1, center=False).sum()[::-1]
    )  # n Flags in interval t to t+window_flags
    # forward-orientation not possible, so right-orientation on reversed data an reverse result

    # Add context information for field+references
    for i in [field] + references:
        df = pd.concat([df, _refCalc(reference=df[i], window_values=window_values)], axis=1)

    # remove rows that contain NAs (new ones occured during predictor calculation)
    df = df.dropna(axis=0, how="any")
    # drop column of automatic flags at time t
    df = df.drop(columns="flag_bin")
    # Load model and predict on df:
    model = joblib.load(path)
    preds = model.predict(df)

    # Get indices of flagged values
    flag_indices = df[preds.astype("bool")].index
    # set Flags
    flagger = flagger.setFlags(field, loc=flag_indices, **kwargs)
    return data, flagger



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import pandas as pd
import numpy as np


def _isValid(data, max_nan_total, max_nan_consec):
    if (max_nan_total is np.inf) & (max_nan_consec is np.inf):
        return True

    nan_mask = data.isna()

    if nan_mask.sum() <= max_nan_total:
        if max_nan_consec is np.inf:
            return True
        elif nan_mask.rolling(window=max_nan_consec + 1).sum().max() <= max_nan_consec:
            return True
        else:
            return False
    else:
        return False


def stdQC(data, max_nan_total=np.inf, max_nan_consec=np.inf):
    """Pandas built in function for statistical moments have quite poor nan- control, so here comes a wrapper that
    will return the standart deviation for a given series input, if the total number of nans in the series does
    not exceed "max_nan_total" and the number of consecutive nans does not exceed max_nan_consec.

    :param data             Pandas Series. The data series, the standart deviation shall be calculated of.
    :param max_nan_total    Integer. Number of np.nan entries allowed to be contained in the series
    :param max_nan_consec   Integer. Maximal number of consecutive nan entries allowed to occure in data.
    """
    if _isValid(data, max_nan_total, max_nan_consec):
        return data.std()
    return np.nan


def varQC(data, max_nan_total=np.inf, max_nan_consec=np.inf):
    """Pandas built in function for statistical moments have quite poor nan- control, so here comes a wrapper that
    will return the variance for a given series input, if the total number of nans in the series does
    not exceed "max_nan_total" and the number of consecutive nans does not exceed max_nan_consec.

    :param data             Pandas Series. The data series, the variance shall be calculated of.
    :param max_nan_total    Integer. Number of np.nan entries allowed to be contained in the series
    :param max_nan_consec   Integer. Maximal number of consecutive nan entries allowed to occure in data.
    """
    if _isValid(data, max_nan_total, max_nan_consec):
        return data.var()
    return np.nan


def meanQC(data, max_nan_total=np.inf, max_nan_consec=np.inf):
    """Pandas built in function for statistical moments have quite poor nan- control, so here comes a wrapper that
    will return the mean for a given series input, if the total number of nans in the series does
    not exceed "max_nan_total" and the number of consecutive nans does not exceed max_nan_consec.

    :param data             Pandas Series. The data series, the mean shall be calculated of.
    :param max_nan_total    Integer. Number of np.nan entries allowed to be contained in the series
    :param max_nan_consec   Integer. Maximal number of consecutive nan entries allowed to occure in data.
    """
    if _isValid(data, max_nan_total, max_nan_consec):
        return data.mean()
    return np.nan



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import ast
from functools import partial
from typing import Any, Dict

import numpy as np
import pandas as pd

from saqc.flagger.baseflagger import BaseFlagger
from saqc.core.config import Params
from saqc.funcs.register import FUNC_MAP
from saqc.core.evaluator.checker import ConfigChecker
from saqc.core.evaluator.transformer import ConfigTransformer


def _dslIsFlagged(flagger, data, flag=None, comparator=None):
    if comparator is None:
        return flagger.isFlagged(data.name, flag=flag)
    return flagger.isFlagged(data.name, flag=flag, comparator=comparator)


def initLocalEnv(data: pd.DataFrame, field: str, flagger: BaseFlagger, nodata: float) -> Dict[str, Any]:

    return {
        "data": data,
        "field": field,
        "this": field,
        "flagger": flagger,
        "NAN": np.nan,
        "NODATA": nodata,
        "GOOD": flagger.GOOD,
        "BAD": flagger.BAD,
        "UNFLAGGED": flagger.UNFLAGGED,
        "ismissing": lambda data: ((data == nodata) | pd.isnull(data)),
        "isflagged": partial(_dslIsFlagged, flagger),
        "abs": np.abs,
        "max": np.nanmax,
        "min": np.nanmin,
        "mean": np.nanmean,
        "sum": np.nansum,
        "std": np.nanstd,
        "len": len,
        "variables": set(data.columns.tolist() + flagger.getFlags().columns.tolist()),
    }


def parseExpression(expr: str) -> ast.AST:
    tree = ast.parse(expr, mode="eval")
    return tree


def compileTree(tree: ast.Expression):
    return compile(ast.fix_missing_locations(tree), "<ast>", mode="eval")


def evalCode(code, global_env=None, local_env=None):
    return eval(code, global_env or {}, local_env or {})


def compileExpression(expr, data, field, flagger, nodata=np.nan):
    local_env = initLocalEnv(data, field, flagger, nodata)
    tree = parseExpression(expr)
    ConfigChecker(local_env, flagger.signature).visit(tree)
    transformed_tree = ConfigTransformer(local_env).visit(tree)
    return local_env, compileTree(transformed_tree)


def evalExpression(expr, data, field, flagger, nodata=np.nan):
    # mask the already flagged value to make all the functions
    # called on the way through the evaluator ignore flagged values
    mask = flagger.isFlagged()
    data_in = data.copy()
    data_in[mask] = np.nan
    local_env, code = compileExpression(expr, data_in, field, flagger, nodata)
    data_result, flagger_result = evalCode(code, FUNC_MAP, local_env)
    # reinject the original values, as we don't want to loose them
    data_result[mask] = data[mask]
    return data_result, flagger_result
